#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  8 13:28:24 2021

@author: deborahkhider

Tests for pyleoclim.core.ui.LipdSeries

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
"""
import pytest
import pyleoclim as pyleo
from urllib.request import urlopen
import json

def importLiPD():
    url = 'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/lipds.json'
    response = urlopen(url)
    d = json.loads(response.read())
    return d

def importEnsLiPD():
    url = 'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/crystalcave_ens.json'
    response = urlopen(url)
    d = json.loads(response.read())
    return d
    
def load_data():
    #Loads stott MD982176 record
    try:
        d = pyleo.Lipd(usr_path='http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004')
    except:
        d = pyleo.Lipd('./example_data/MD982176.Stott.2004.lpd')
    return d
    
def get_ts():
    d=load_data()
    ts=d.to_LipdSeries(number=3)
    return ts

class TestUiLipdSeriesMap():
    ''' test LipdSeries.map()
    '''
    
    def test_map_t0(self):
        ts=get_ts()
        fig,ax=ts.map()
        pyleo.closefig(fig)

class TestUiLipdSeriesgetMetadata():
    ''' test LipdSeries.getMetadata
    '''
    
    def test_getMetadata_t0(self):
        ts=get_ts()
        res=ts.getMetadata()

class TestUiLipdSeriesDashboard():
    ''' test LipdSeries.Dashboard
    '''
    
    def test_dashboard_t0(self):
        ts=get_ts()
        fig,ax = ts.dashboard()
        pyleo.closefig(fig)

class TestUiLipdSeriesMapNearRecord():
    '''Test LipdSeries.MapNear Record
    
    Requires a dictionary of LiPDs and selection
    '''
    
    def test_mapNearRecord_t0(self):
        D=importLiPD()
        d=pyleo.Lipd(lipd_dict=D)
        ts = d.to_LipdSeries(number=6)
        res=ts.mapNearRecord(d)
        pyleo.closefig(res[0])
    
    def test_mapNearRecord_t1(self):
        D=importLiPD()
        d=pyleo.Lipd(lipd_dict=D)
        ts = d.to_LipdSeries(number=6)
        res=ts.mapNearRecord(d,n=6)
        pyleo.closefig(res[0])
    
    def test_mapNearRecord_t2(self):
        D=importLiPD()
        d=pyleo.Lipd(lipd_dict=D)
        ts = d.to_LipdSeries(number=6)
        res=ts.mapNearRecord(d,radius=1000)
        pyleo.closefig(res[0])

class TestUiLipdSeriesChronEnsembleToPaleo():
    ''' Test the ability to get the chron ensemble tables
    '''

    def test_chronEnsembletoPaleo_t0(self):
        D=importEnsLiPD()
        d=pyleo.Lipd(lipd_dict=D)
        ts = d.to_LipdSeries(number=2)
        ens = ts.chronEnsembleToPaleo(d)
        
        assert type(ens)==pyleo.core.EnsembleSeries
        
        
class TestUiLipdSeriesPlotAgeDepth():
    '''test LipdSeries.plot_age_depth
    '''

    def test_plot_age_depth_t0(self):
        D=importEnsLiPD()
        d=pyleo.Lipd(lipd_dict=D)
        ts = d.to_LipdSeries(number=2)
        fig,ax = ts.plot_age_depth()
        pyleo.closefig(fig)
    
    @pytest.mark.parametrize('traces', [10,0])
    def test_plot_age_depth_t1(self,traces):
        D=importEnsLiPD()
        d=pyleo.Lipd(lipd_dict=D)
        ts = d.to_LipdSeries(number=2)
        fig,ax = ts.plot_age_depth(ensemble=True, D=d, num_traces=traces)
        pyleo.closefig(fig)
        


"""
A MultipleSeries object is a collection (more precisely, a 
list) of multiple Series objects. This is handy in case you want to apply the same method 
to such a collection at once (e.g. process a bunch of series in a consistent fashion).
"""

from ..utils import tsutils, plotting
from ..utils import correlation as corrutils

from ..core.correns import CorrEns
from ..core.scalograms import MultipleScalogram
from ..core.psds import MultiplePSD
from ..core.spatialdecomp import SpatialDecomp

import numpy as np
from copy import deepcopy

from matplotlib.ticker import FormatStrFormatter
import matplotlib.transforms as transforms
import matplotlib as mpl
import matplotlib.pyplot as plt

from tqdm import tqdm
from scipy import stats
from statsmodels.multivariate.pca import PCA

class MultipleSeries:
    '''MultipleSeries object.

    This object handles a collection of the type Series and can be created from a list of such objects.
    MultipleSeries should be used when the need to run analysis on multiple records arises, such as running principal component analysis.
    Some of the methods automatically transform the time axis prior to analysis to ensure consistency.

    Parameters
    ----------

    series_list : list
    
        a list of pyleoclim.Series objects

    time_unit : str
    
        The target time unit for every series in the list.
        If None, then no conversion will be applied;
        Otherwise, the time unit of every series in the list will be converted to the target.

    name : str
   
        name of the collection of timeseries (e.g. 'PAGES 2k ice cores')

    Examples
    --------
    .. ipython:: python
        :okwarning:
        :okexcept:

        import pyleoclim as pyleo
        import pandas as pd
        data = pd.read_csv(
            'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
            skiprows=0, header=1
        )
        time = data.iloc[:,1]
        value = data.iloc[:,2]
        ts1 = pyleo.Series(time=time, value=value, time_unit='years')
        ts2 = pyleo.Series(time=time, value=value, time_unit='years')
        ms = pyleo.MultipleSeries([ts1, ts2], name = 'SOI x2')
    '''
    def __init__(self, series_list, time_unit=None, name=None):
        self.series_list = series_list
        self.time_unit = time_unit
        self.name = name

        if self.time_unit is not None:
            new_ts_list = []
            for ts in self.series_list:
                new_ts = ts.convert_time_unit(time_unit=self.time_unit)
                new_ts_list.append(new_ts)

            self.series_list = new_ts_list

    def convert_time_unit(self, time_unit='years'):
        ''' Convert the time units of the object

        Parameters
        ----------

        time_unit : str
        
            the target time unit, possible input:
            {
                'year', 'years', 'yr', 'yrs',
                'y BP', 'yr BP', 'yrs BP', 'year BP', 'years BP',
                'ky BP', 'kyr BP', 'kyrs BP', 'ka BP', 'ka',
                'my BP', 'myr BP', 'myrs BP', 'ma BP', 'ma',
            }

        Examples
        --------
        
        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            data = pd.read_csv(
                'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
                skiprows=0, header=1
            )
            time = data.iloc[:,1]
            value = data.iloc[:,2]
            ts1 = pyleo.Series(time=time, value=value, time_unit='years')
            ts2 = pyleo.Series(time=time, value=value, time_unit='years')
            ms = pyleo.MultipleSeries([ts1, ts2])
            new_ms = ms.convert_time_unit('yr BP')
            print('Original timeseries:')
            print('time unit:', ms.time_unit)
            print()
            print('Converted timeseries:')
            print('time unit:', new_ms.time_unit)
        '''

        new_ms = self.copy()
        new_ts_list = []
        for ts in self.series_list:
            new_ts = ts.convert_time_unit(time_unit=time_unit)
            new_ts_list.append(new_ts)

        new_ms.time_unit = time_unit
        new_ms.series_list = new_ts_list
        return new_ms

    def filter(self, cutoff_freq=None, cutoff_scale=None, method='butterworth', **kwargs):
        ''' Filtering the timeseries in the MultipleSeries object

        Parameters
        ----------

        method : str; {'savitzky-golay', 'butterworth', 'firwin', 'lanczos'}
        
            The filtering method  
            - 'butterworth': the Butterworth method (default)
            - 'savitzky-golay': the Savitzky-Golay method
            - 'firwin': FIR filter design using the window method, with default window as Hamming
            - 'lanczos': lowpass filter via Lanczos resampling

        cutoff_freq : float or list
        
            The cutoff frequency only works with the Butterworth method.
            If a float, it is interpreted as a low-frequency cutoff (lowpass).
            If a list,  it is interpreted as a frequency band (f1, f2), with f1 < f2 (bandpass).

        cutoff_scale : float or list
        
            cutoff_freq = 1 / cutoff_scale
            The cutoff scale only works with the Butterworth method and when cutoff_freq is None.
            If a float, it is interpreted as a low-frequency (high-scale) cutoff (lowpass).
            If a list,  it is interpreted as a frequency band (f1, f2), with f1 < f2 (bandpass).

        kwargs : dict
        
            A dictionary of the keyword arguments for the filtering method,
            See pyleoclim.utils.filter.savitzky_golay, pyleoclim.utils.filter.butterworth, pyleoclim.utils.filter.firwin, and pyleoclim.utils.filter.lanczos for the details

        Returns
        -------

        ms : MultipleSeries

        See also
        --------

        pyleoclim.series.Series.filter : filtering for Series objects    

        pyleoclim.utils.filter.butterworth : Butterworth method

        pyleoclim.utils.filter.savitzky_golay : Savitzky-Golay method

        pyleoclim.utils.filter.firwin : FIR filter design using the window method

        pyleoclim.utils.filter.lanczos : lowpass filter via Lanczos resampling

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            data = pd.read_csv(
                'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
                skiprows=0, header=1
            )
            time = data.iloc[:,1]
            value = data.iloc[:,2]
            ts1 = pyleo.Series(time=time, value=value, time_unit='years')
            ts2 = pyleo.Series(time=time, value=value, time_unit='years')
            ms = pyleo.MultipleSeries([ts1, ts2], name = 'SOI x2')
            ms_filter = ms.filter(method='lanczos',cutoff_scale=20)
        '''

        ms = self.copy()

        new_tslist = []

        for ts in self.series_list:
            new_tslist.append(ts.filter(cutoff_freq=cutoff_freq, cutoff_scale=cutoff_scale, method=method, **kwargs))

        ms.series_list = new_tslist

        return ms

    def append(self,ts):
        '''Append timeseries ts to MultipleSeries object

        Parameters
        ----------

        ts : pyleoclim.Series
        
            The pyleoclim Series object to be appended to the MultipleSeries object

        Returns
        -------

        ms : MultipleSeries
        
            The augmented object, comprising the old one plus `ts`

        See also
        --------

        pyleoclim.core.series.Series : A Pyleoclim Series object

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            data = pd.read_csv(
                'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
                skiprows=0, header=1
            )
            time = data.iloc[:,1]
            value = data.iloc[:,2]
            ts1 = pyleo.Series(time=time, value=value, time_unit='years')
            ts2 = pyleo.Series(time=time, value=value, time_unit='years')
            ms = pyleo.MultipleSeries([ts1], name = 'SOI x2')
            ms.append(ts2)
        '''
        ms = self.copy()
        ts_list = deepcopy(ms.series_list)
        ts_list.append(ts)
        ms = MultipleSeries(ts_list)
        return ms

    def copy(self):
        '''Copy the object

        Returns
        -------

        ms : MultipleSeries
        
            The copied version of the pyleoclim.MultipleSeries object

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            data = pd.read_csv(
                'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
                skiprows=0, header=1
            )
            time = data.iloc[:,1]
            value = data.iloc[:,2]
            ts1 = pyleo.Series(time=time, value=value, time_unit='years')
            ts2 = pyleo.Series(time=time, value=value, time_unit='years')
            ms = pyleo.MultipleSeries([ts1], name = 'SOI x2')
            ms_copy = ms.copy()
        '''
        return deepcopy(self)
    
    def flip(self, axis='value'):
        '''
        Flips the Series along one or both axes

        Parameters
        ----------
        axis : str, optional
            The axis along which the Series will be flipped. The default is 'value'.
            Other acceptable options are 'time' or 'both'.
            TODO: enable time flipping after paleopandas is released
            
        Returns
        -------
        ms : MultipleSeries
            The flipped object
            
         Examples
         --------

         .. ipython:: python
             :okwarning:
             :okexcept:

             import pyleoclim as pyleo
             url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
             data = pyleo.Lipd(usr_path = url)
             tslist = data.to_LipdSeriesList()
             tslist = tslist[2:] # drop the first two series which only concerns age and depth
             ms = pyleo.MultipleSeries(tslist)
             
             @savefig ms_flip.png
             fig, ax = ms.flip().stackplot()
             pyleo.closefig(fig) 
            
        
        Note that labels have been updated to reflect the flip
        '''
        
        ms=self.copy()
        for idx,item in enumerate(ms.series_list):
            s=item.flip(axis=axis, keep_log=False)
            ms.series_list[idx]=s
        
        return ms

    def standardize(self):
        '''Standardize each series object in a collection

        Returns
        -------

        ms : MultipleSeries
        
            The standardized pyleoclim.MultipleSeries object

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            data = pd.read_csv(
                'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
                skiprows=0, header=1
            )
            time = data.iloc[:,1]
            value = data.iloc[:,2]
            ts1 = pyleo.Series(time=time, value=value, time_unit='years')
            ts2 = pyleo.Series(time=time, value=value, time_unit='years')
            ms = pyleo.MultipleSeries([ts1], name = 'SOI x2')
            ms_std = ms.standardize()
        '''
        ms=self.copy()
        for idx,item in enumerate(ms.series_list):
            s=item.copy()
            v_mod=tsutils.standardize(item.value)[0]
            s.value=v_mod
            ms.series_list[idx]=s
        return ms

    def increments(self, step_style='median', verbose=False):
        '''
        Extract grid properties (start, stop, step) of all the Series objects in a collection.

        Parameters
        ----------
        
        step_style : str; {'median','mean','mode','max'}
        
            Method to obtain a representative step if x is not evenly spaced.
            Valid entries: 'median' [default], 'mean', 'mode' or 'max'.
            The "mode" is the most frequent entry in a dataset, and may be a good choice if the timeseries
            is nearly equally spaced but for a few gaps.

            "max" is a conservative choice, appropriate for binning methods and Gaussian kernel coarse-graining
        
        verbose : bool
            If True, will print out warning messages when they appear
            
        Returns
        -------

        increments : numpy.array
        
            n x 3 array, where n is the number of series,
            
            * index 0 is the earliest time among all Series
            
            * index 1 is the latest time among all Series
            
            * index 2 is the step, chosen according to step_style
            
        See also
        --------
        
        pyleoclim.utils.tsutils.increments :  underlying array-level utility

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            data = pd.read_csv(
                'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
                skiprows=0, header=1
            )
            time = data.iloc[:,1]
            value = data.iloc[:,2]
            ts1 = pyleo.Series(time=time, value=value, time_unit='years')
            ts2 = pyleo.Series(time=time, value=value, time_unit='years')
            ms = pyleo.MultipleSeries([ts1], name = 'SOI x2')
            increments = ms.increments()

        '''
        gp = np.empty((len(self.series_list),3)) # obtain grid parameters
        for idx,item in enumerate(self.series_list):
            item      = item.clean(verbose=verbose)
            gp[idx,:] = tsutils.increments(item.time, step_style=step_style)

        return gp

    def common_time(self, method='interp', step = None, start = None, stop = None, step_style = None, **kwargs):
        ''' Aligns the time axes of a MultipleSeries object
        
        The alignment is achieved via binning, interpolation, or Gaussian kernel. Alignment is critical for workflows
        that need to assume a common time axis for the group of series under consideration.

        The common time axis is characterized by the following parameters:

        start : the latest start date of the bunch (maximun of the minima)

        stop  : the earliest stop date of the bunch (minimum of the maxima)

        step  : The representative spacing between consecutive values

        Optional arguments for binning, Gaussian kernel (gkernel) interpolation are those of the underling functions.

        If any of the time axes are retrograde, this step makes them prograde.

        Parameters
        ----------
        
        method :  string; {'bin','interp','gkernel'}
        
            either 'bin', 'interp' [default] or 'gkernel'
            
        step : float
        
            common step for all time axes. Default is None and inferred from the timeseries spacing
            
        start : float
        
            starting point of the common time axis. Default is None and inferred as the max of the min of the time axes for the timeseries.
            
        stop : float
        
            end point of the common time axis. Default is None and inferred as the min of the max of the time axes for the timeseries.
        
        step_style : string; {'median', 'mean', 'mode', 'max'}
        
            Method to obtain a representative step among all Series (using tsutils.increments).
            Default value is None, so that it will be chosen according to the method: 'max' for bin and gkernel, 'mean' for interp. 

        kwargs: dict
        
            keyword arguments (dictionary) of the bin, gkernel or interp methods

        Returns
        -------

        ms : MultipleSeries
        
            The MultipleSeries objects with all series aligned to the same time axis.


        See also
        --------

        pyleoclim.utils.tsutils.bin : put timeseries values into bins of equal size (possibly leaving NaNs in).

        pyleoclim.utils.tsutils.gkernel : coarse-graining using a Gaussian kernel

        pyleoclim.utils.tsutils.interp : interpolation onto a regular grid (default = linear interpolation)

        pyleoclim.utils.tsutils.increments : infer grid properties

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import numpy as np
            import pyleoclim as pyleo
            import matplotlib.pyplot as plt
            from pyleoclim.utils.tsmodel import colored_noise

            # create 2 incompletely sampled series
            ns = 2 ; nt = 200; n_del = 20
            serieslist = []

            for j in range(ns):
                t = np.arange(nt)
                v = colored_noise(alpha=1, t=t)
                deleted_idx = np.random.choice(range(np.size(t)), n_del, replace=False)
                tu =  np.delete(t, deleted_idx)
                vu =  np.delete(v, deleted_idx)
                ts = pyleo.Series(time = tu, value = vu, label = 'series ' + str(j+1))
                serieslist.append(ts)

            # create MS object from the list
            ms = pyleo.MultipleSeries(serieslist)

            @savefig ms_common_time.png
            fig, ax = plt.subplots(2,2,sharex=True,sharey=True, figsize=(10,8))
            ax = ax.flatten()
            # apply common_time with default parameters
            msc = ms.common_time()
            msc.plot(title='linear interpolation',ax=ax[0], legend=False)

            # apply common_time with binning
            msc = ms.common_time(method='bin')
            msc.plot(title='Binning',ax=ax[1], legend=False)

            # apply common_time with gkernel
            msc = ms.common_time(method='gkernel')
            msc.plot(title=r'Gaussian kernel ($h=3$)',ax=ax[2],legend=False)

            # apply common_time with gkernel and a large bandwidth
            msc = ms.common_time(method='gkernel', h=.5)
            msc.plot(title=r'Gaussian kernel ($h=.5$)',ax=ax[3],legend=False)
            fig.tight_layout()
            # Optional close fig after plotting
            pyleo.closefig(fig)
        '''
        
        # specify stepping style
        if step_style == None: # if step style isn't specified, pick a robust choice according to method
            if method == 'bin' or method == 'gkernel':
                step_style = 'max'
            elif  method == 'interp':
                step_style = 'mean'
               
        # obtain grid properties with given step_style
        gp = self.increments(step_style=step_style)  
        
        # define grid step     
        if step is not None and step > 0:
            common_step = step 
        else:
            if step_style == 'mean':
                common_step = gp[:,2].mean()
            elif step_style == 'max':
                common_step = gp[:,2].max()
            elif step_style == 'mode':
                common_step = stats.mode(gp[:,2])[0][0]
            else:
                common_step = np.median(gp[:,2])
   
        # define start and stop
        if start is None: 
            start = gp[:,0].max() # pick the latest of the start times
        
        if stop is None:
            stop  = gp[:,1].min() # pick the earliest of the stop times
        
        if start > stop:
            raise ValueError('At least one series has no common time interval with others. Please check the time axis of the series.')
    
        ms = self.copy()

        # apply each method
        if method == 'bin':
            for idx,item in enumerate(self.series_list):
                ts = item.copy()
                d = tsutils.bin(ts.time, ts.value, bin_size=common_step, start=start, stop=stop, evenly_spaced = False, **kwargs)
                ts.time  = d['bins']
                ts.value = d['binned_values']
                ms.series_list[idx] = ts

        elif method == 'interp':
            for idx,item in enumerate(self.series_list):
                ts = item.copy()
                ti, vi = tsutils.interp(ts.time, ts.value, step=common_step, start=start, stop=stop,**kwargs)
                ts.time  = ti
                ts.value = vi
                ms.series_list[idx] = ts

        elif method == 'gkernel':
            for idx,item in enumerate(self.series_list):
                ts = item.copy()
                ti, vi = tsutils.gkernel(ts.time,ts.value,step=common_step, start=start, stop=stop, **kwargs)
                ts.time  = ti
                ts.value = vi
                ms.series_list[idx] = ts.clean() # remove NaNs

        else:
            raise NameError('Unknown methods; no action taken')

        return ms

    def correlation(self, target=None, timespan=None, alpha=0.05, settings=None, 
                    fdr_kwargs=None, common_time_kwargs=None, mute_pbar=False, seed=None):
        ''' Calculate the correlation between a MultipleSeries and a target Series

        Parameters
        ----------

        target : pyleoclim.Series, optional
        
            The Series against which to take the correlation. If the target Series is not specified, then the 1st member of MultipleSeries will be used as the target

        timespan : tuple, optional
        
            The time interval over which to perform the calculation

        alpha : float
        
            The significance level (0.05 by default)

        settings : dict
        
            Parameters for the correlation function, including:

            nsim : int
                the number of simulations (default: 1000)
            method : str, {'ttest','isopersistent','isospectral' (default)}
                method for significance testing

        fdr_kwargs : dict
        
            Parameters for the FDR function

        common_time_kwargs : dict
        
            Parameters for the method MultipleSeries.common_time()

        mute_pbar : bool; {True,False}
        
            If True, the progressbar will be muted. Default is False.
        
        seed : float or int
        
            random seed for isopersistent and isospectral methods

        Returns
        -------

        corr : CorrEns
        
            the result object

        See also
        --------

        pyleoclim.utils.correlation.corr_sig : Correlation function

        pyleoclim.utils.correlation.fdr : FDR function
        
        pyleoclim.core.correns.CorrEns : the correlation ensemble object

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            from pyleoclim.utils.tsmodel import colored_noise
            import numpy as np

            nt = 100
            t0 = np.arange(nt)
            v0 = colored_noise(alpha=1, t=t0)
            noise = np.random.normal(loc=0, scale=1, size=nt)

            ts0 = pyleo.Series(time=t0, value=v0)
            ts1 = pyleo.Series(time=t0, value=v0+noise)
            ts2 = pyleo.Series(time=t0, value=v0+2*noise)
            ts3 = pyleo.Series(time=t0, value=v0+1/2*noise)

            ts_list = [ts1, ts2, ts3]

            ms = pyleo.MultipleSeries(ts_list)
            ts_target = ts0

        Correlation between the MultipleSeries object and a target Series. We also set an arbitrary random seed to ensure reproducibility:
       
        .. ipython:: python
            :okwarning:
            :okexcept:
           
            corr_res = ms.correlation(ts_target, settings={'nsim': 20}, seed=2333)
            print(corr_res)
        
        Correlation among the series of the MultipleSeries object
        
        .. ipython:: python
            :okwarning:
            :okexcept:

            corr_res = ms.correlation(settings={'nsim': 20}, seed=2333)
            print(corr_res)

        '''
        r_list = []
        signif_list = []
        p_list = []

        if target is None:
            target = self.series_list[0]

        print("Looping over "+ str(len(self.series_list)) +" Series in collection")
        for idx, ts in tqdm(enumerate(self.series_list),  total=len(self.series_list), disable=mute_pbar):
            corr_res = ts.correlation(target, timespan=timespan, alpha=alpha, settings=settings, common_time_kwargs=common_time_kwargs, seed=seed)
            r_list.append(corr_res.r)
            signif_list.append(corr_res.signif)
            p_list.append(corr_res.p)

        r_list = np.array(r_list)
        signif_fdr_list = []
        fdr_kwargs = {} if fdr_kwargs is None else fdr_kwargs.copy()
        args = {}
        args.update(fdr_kwargs)
        for i in range(np.size(signif_list)):
            signif_fdr_list.append(False)

        fdr_res = corrutils.fdr(p_list, **fdr_kwargs)
        if fdr_res is not None:
            for i in fdr_res:
                signif_fdr_list[i] = True

        corr_ens = CorrEns(r_list, p_list, signif_list, signif_fdr_list, alpha)
        return corr_ens


    def equal_lengths(self):
        ''' Test whether all series in object have equal length

        Returns
        -------

        flag : bool
        
            Whether or not the Series in the pyleo.MultipleSeries object are of equal length

        lengths : list 
        
            List of the lengths of the series in object
            
        See also
        --------
        
        pyleoclim.core.multipleseries.MultipleSeries.common_time : Aligns the time axes of a MultipleSeries object
            
        Examples
        --------    

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            import pandas as pd
            data = pd.read_csv(
                'https://raw.githubusercontent.com/LinkedEarth/Pyleoclim_util/Development/example_data/soi_data.csv',
                skiprows=0, header=1
            )
            time = data.iloc[:,1]
            value = data.iloc[:,2]
            ts1 = pyleo.Series(time=time, value=value, time_unit='years')
            ts2 = pyleo.Series(time=time, value=value, time_unit='years')
            ms = pyleo.MultipleSeries([ts1], name = 'SOI x2')
            flag, lengths = ms.equal_lengths()
            print(flag)
        '''

        lengths = []
        for ts in self.series_list:
            lengths.append(len(ts.value))

        L = lengths[0]
        r = lengths[1:]
        flag = all (l==L for l in r)

        return flag, lengths

    def pca(self,weights=None,missing='fill-em',tol_em=5e-03, max_em_iter=100,**pca_kwargs):
        '''Principal Component Analysis (Empirical Orthogonal Functions)

        Decomposition of dataset ys in terms of orthogonal basis functions.
        Tolerant to missing values, infilled by an EM algorithm.

        Do make sure the time axes are aligned, however! (e.g. use `common_time()`)

        Algorithm from statsmodels: https://www.statsmodels.org/stable/generated/statsmodels.multivariate.pca.PCA.html

        Parameters
        ----------

        weights : ndarray, optional
        
            Series weights to use after transforming data according to standardize
            or demean when computing the principal components.

        missing : {str, None}
        
            Method for missing data.  Choices are:

            * 'drop-row' - drop rows with missing values.
            * 'drop-col' - drop columns with missing values.
            * 'drop-min' - drop either rows or columns, choosing by data retention.
            * 'fill-em' - use EM algorithm to fill missing value.  ncomp should be
              set to the number of factors required.
            * `None` raises if data contains NaN values.

        tol_em : float
        
            Tolerance to use when checking for convergence of the EM algorithm.
            
        max_em_iter : int
        
            Maximum iterations for the EM algorithm.

        Returns
        -------

        res: SpatialDecomp

            Resulting pyleoclim.SpatialDecomp object
        
        See also
        --------
        
        pyleoclim.utils.tsutils.eff_sample_size : Effective Sample Size of timeseries y

        pyleoclim.core.spatialdecomp.SpatialDecomp : The spatial decomposition object
        
        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            tslist = data.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist).common_time()

            res = ms.pca() # carry out PCA

            @savefig ms_pca1.png
            fig1, ax1 = res.screeplot() # plot the eigenvalue spectrum
            pyleo.closefig(fig1)    # Optional close fig after plotting

            @savefig ms_pca2.png
            fig2, ax2 = res.modeplot() # plot the first mode
            pyleo.closefig(fig2)    # Optional close fig after plotting
        '''
        flag, lengths = self.equal_lengths()

        if flag==False:
            print('All Time Series should be of same length. Apply common_time() first')
        else: # if all series have equal length
            p = len(lengths)
            n = lengths[0]
            ys = np.empty((n,p))
            for j in range(p):
                ys[:,j] = self.series_list[j].value  # fill in data matrix

        nc = min(ys.shape) # number of components to return

        out  = PCA(ys,weights=weights,missing=missing,tol_em=tol_em, max_em_iter=max_em_iter,**pca_kwargs)

        # compute effective sample size
        PC1  = out.factors[:,0]
        neff = tsutils.eff_sample_size(PC1)

        # compute percent variance
        pctvar = out.eigenvals**2/np.sum(out.eigenvals**2)*100

        # assign result to SpatiamDecomp class
        # Note: need to grab coordinates from Series or LiPDSeries
        res = SpatialDecomp(name='PCA', time = self.series_list[0].time, neff= neff,
                            pcs = out.scores, pctvar = pctvar,  locs = None,
                            eigvals = out.eigenvals, eigvecs = out.eigenvecs)
        return res

    # def mcpca(self,nMC=200,**pca_kwargs):
    #     ''' Monte Carlo Principal Component Analysis

    #     (UNDER REPAIR)

    #     Parameters
    #     ----------

    #     nMC : int
    #         number of Monte Carlo simulations

    #     pca_kwargs : tuple


    #     Returns
    #     -------
    #     res : dictionary containing:

    #         - eigval : eigenvalues (nrec,)
    #         - eig_ar1 : eigenvalues of the AR(1) ensemble (nrec, nMC)
    #         - pcs  : PC series of all components (nrec, nt)
    #         - eofs : EOFs of all components (nrec, nrec)

    #     References:
    #     ----------
    #     Deininger, M., McDermott, F., Mudelsee, M. et al. (2017): Coherency of late Holocene
    #     European speleothem δ18O records linked to North Atlantic Ocean circulation.
    #     Climate Dynamics, 49, 595–618. https://doi.org/10.1007/s00382-016-3360-8

    #     See also
    #     --------

    #     pyleoclim.utils.decomposition.mcpca: Monte Carlo PCA

    #     Examples
    #     --------

    #     .. ipython:: python
    #         :okwarning:

    #         import pyleoclim as pyleo
    #         url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
    #         data = pyleo.Lipd(usr_path = url)
    #         tslist = data.to_LipdSeriesList()
    #         tslist = tslist[2:] # drop the first two series which only concerns age and depth
    #         ms = pyleo.MultipleSeries(tslist)

    #         # msc = ms.common_time()

    #         # res = msc.pca(nMC=20)

    #     '''
    #     flag, lengths = self.equal_lengths()

    #     if flag==False:
    #         print('All Time Series should be of same length. Apply common_time() first')
    #     else: # if all series have equal length
    #         p = len(lengths)
    #         n = lengths[0]
    #         ys = np.empty((n,p))
    #         for j in range(p):
    #             ys[:,j] = self.series_list[j].value

    #     res = decomposition.mcpca(ys, nMC, **pca_kwargs)
    #     return res

    def bin(self, **kwargs):
        '''Aligns the time axes of a MultipleSeries object, via binning.

        This is critical for workflows that need to assume a common time axis
        for the group of series under consideration.

        The common time axis is characterized by the following parameters:

        start : the latest start date of the bunch (maximin of the minima)

        stop  : the earliest stop date of the bunch (minimum of the maxima)

        step  : The representative spacing between consecutive values (mean of the median spacings)

        This is a special case of the common_time function.

        Parameters
        ----------

        kwargs : dict
        
            Arguments for the binning function. See pyleoclim.utils.tsutils.bin

        Returns
        -------

        ms : MultipleSeries
        
            The MultipleSeries objects with all series aligned to the same time axis.

        See also
        --------

        pyleoclim.core.multipleseries.MultipleSeries.common_time: Base function on which this operates

        pyleoclim.utils.tsutils.bin: Underlying binning function

        pyleoclim.core.series.Series.bin: Bin function for Series object

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            tslist = data.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)
            msbin = ms.bin()
        '''

        ms = self.copy()

        ms = ms.common_time(method = 'bin', **kwargs)

        return ms

    def gkernel(self, **kwargs):
        ''' Aligns the time axes of a MultipleSeries object, via Gaussian kernel.
        
        This is critical for workflows that need to assume a common time axis
        for the group of series under consideration.

        The common time axis is characterized by the following parameters:

        start : the latest start date of the bunch (maximin of the minima)

        stop  : the earliest stop date of the bunch (minimum of the maxima)

        step  : The representative spacing between consecutive values (mean of the median spacings)

        This is a special case of the common_time function.

        Parameters
        ----------

        kwargs : dict
        
            Arguments for gkernel. See pyleoclim.utils.tsutils.gkernel for details.

        Returns
        -------

        ms : MultipleSeries
            The MultipleSeries objects with all series aligned to the same time axis.

        See also
        --------

        pyleoclim.core.multipleseries.MultipleSeries.common_time: Base function on which this operates

        pyleoclim.utils.tsutils.gkernel: Underlying kernel module

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            tslist = data.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)
            msk = ms.gkernel()

        '''

        ms = self.copy()

        ms = ms.common_time(method = 'gkernel', **kwargs)

        return ms

    def interp(self, **kwargs):
        ''' Aligns the time axes of a MultipleSeries object, via interpolation.
        
        This is critical for workflows that need to assume a common time axis
        for the group of series under consideration.


        The common time axis is characterized by the following parameters:

        start : the latest start date of the bunch (maximin of the minima)

        stop  : the earliest stop date of the bunch (minimum of the maxima)

        step  : The representative spacing between consecutive values (mean of the median spacings)

        This is a special case of the common_time function.

        Parameters
        ----------

        kwargs: keyword arguments (dictionary) for the interpolation method

        Returns
        -------

        ms : MultipleSeries
        
            The MultipleSeries objects with all series aligned to the same time axis.

        See also
        --------

        pyleoclim.core.multipleseries.MultipleSeries.common_time: Base function on which this operates

        pyleoclim.utils.tsutils.interp: Underlying interpolation function

        pyleoclim.core.series.Series.interp: Interpolation function for Series object

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            tslist = data.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)
            msinterp = ms.interp()

        '''
        ms = self.copy()

        ms = ms.common_time(method='interp', **kwargs)

        return ms

    def detrend(self,method='emd',**kwargs):
        '''Detrend timeseries

        Parameters
        ----------
        
        method : str, optional
        
            The method for detrending. The default is 'emd'.
            Options include:
                * linear: the result of a linear least-squares fit to y is subtracted from y.
                * constant: only the mean of data is subtrated.
                * 'savitzky-golay', y is filtered using the Savitzky-Golay filters and the resulting filtered series is subtracted from y.
                * 'emd' (default): Empirical mode decomposition. The last mode is assumed to be the trend and removed from the series
                
        **kwargs : dict
            Relevant arguments for each of the methods.

        Returns
        -------
        
        ms : MultipleSeries
        
            The detrended timeseries

        See also
        --------

        pyleoclim.core.series.Series.detrend : Detrending for a single series

        pyleoclim.utils.tsutils.detrend : Detrending function
        '''
        ms=self.copy()
        for idx,item in enumerate(ms.series_list):
            s=item.copy()
            v_mod, _=tsutils.detrend(item.value,x=item.time,method=method,**kwargs)
            s.value=v_mod
            ms.series_list[idx]=s
        return ms

    def spectral(self, method='lomb_scargle', settings=None, mute_pbar=False, freq_method='log', 
                freq_kwargs=None, label=None, verbose=False, scalogram_list=None):
        ''' Perform spectral analysis on the timeseries

        Parameters
        ----------

        method : str; {'wwz', 'mtm', 'lomb_scargle', 'welch', 'periodogram', 'cwt'}

        freq_method : str; {'log','scale', 'nfft', 'lomb_scargle', 'welch'}

        freq_kwargs : dict
        
            Arguments for frequency vector

        settings : dict
        
            Arguments for the specific spectral method

        label : str
        
            Label for the PSD object

        verbose : bool
        
            If True, will print warning messages if there is any

        mute_pbar : bool
        
            Mute the progress bar. Default is False.

        scalogram_list : pyleoclim.MultipleScalogram
        
            Multiple scalogram object containing pre-computed scalograms to use when calculating spectra, only works with wwz or cwt

        Returns
        -------

        psd : MultiplePSD
        
            A Multiple PSD object

        See also
        --------
        
        pyleoclim.utils.spectral.mtm : Spectral analysis using the Multitaper approach

        pyleoclim.utils.spectral.lomb_scargle : Spectral analysis using the Lomb-Scargle method

        pyleoclim.utils.spectral.welch: Spectral analysis using the Welch segement approach

        pyleoclim.utils.spectral.periodogram: Spectral anaysis using the basic Fourier transform

        pyleoclim.utils.spectral.wwz_psd : Spectral analysis using the Wavelet Weighted Z transform

        pyleoclim.utils.spectral.cwt_psd : Spectral analysis using the continuous Wavelet Transform as implemented by Torrence and Compo

        pyleoclim.utils.wavelet.make_freq_vector : Functions to create the frequency vector

        pyleoclim.utils.tsutils.detrend : Detrending function

        pyleoclim.core.series.Series.spectral : Spectral analysis for a single timeseries

        pyleoclim.core.PSD.PSD : PSD object

        pyleoclim.core.psds.MultiplePSD : Multiple PSD object

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            tslist = data.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)
            ms_psd = ms.spectral()
        '''
        settings = {} if settings is None else settings.copy()

        psd_list = []
        if method in ['wwz','cwt'] and scalogram_list:
            scalogram_list_len = len(scalogram_list.scalogram_list)
            series_len = len(self.series_list)

            #In the case where the scalogram list and series list are the same we can re-use scalograms in a one to one fashion
            #OR if the scalogram list is longer than the series list we use as many scalograms from the scalogram list as we need
            if scalogram_list_len >= series_len:
                for idx, s in enumerate(tqdm(self.series_list, desc='Performing spectral analysis on individual series', position=0, leave=True, disable=mute_pbar)):
                    psd_tmp = s.spectral(method=method, settings=settings, freq_method=freq_method, freq_kwargs=freq_kwargs, label=label, verbose=verbose,scalogram = scalogram_list.scalogram_list[idx])
                    psd_list.append(psd_tmp)
            #If the scalogram list isn't as long as the series list, we re-use all the scalograms we can and then calculate the rest
            elif scalogram_list_len < series_len:
                for idx, s in enumerate(tqdm(self.series_list, desc='Performing spectral analysis on individual series', position=0, leave=True, disable=mute_pbar)):
                    if idx < scalogram_list_len:
                        psd_tmp = s.spectral(method=method, settings=settings, freq_method=freq_method, freq_kwargs=freq_kwargs, label=label, verbose=verbose,scalogram = scalogram_list.scalogram_list[idx])
                        psd_list.append(psd_tmp)
                    else:
                        psd_tmp = s.spectral(method=method, settings=settings, freq_method=freq_method, freq_kwargs=freq_kwargs, label=label, verbose=verbose)
                        psd_list.append(psd_tmp)
        else:
            for s in tqdm(self.series_list, desc='Performing spectral analysis on individual series', position=0, leave=True, disable=mute_pbar):
                psd_tmp = s.spectral(method=method, settings=settings, freq_method=freq_method, freq_kwargs=freq_kwargs, label=label, verbose=verbose)
                psd_list.append(psd_tmp)

        psds = MultiplePSD(psd_list=psd_list)

        return psds

    def wavelet(self, method='cwt', settings={}, freq_method='log', freq_kwargs=None, verbose=False, mute_pbar=False):
        '''Wavelet analysis

        Parameters
        ----------
        
        method : str {wwz, cwt}
        
            - cwt - the continuous wavelet transform (as per Torrence and Compo [1998])
                is appropriate only for evenly-spaced series.
            - wwz - the weighted wavelet Z-transform (as per Foster [1996])
                is appropriate for both evenly and unevenly-spaced series.

            Default is cwt, returning an error if the Series is unevenly-spaced.

        settings : dict, optional
        
            Settings for the particular method. The default is {}.

        freq_method : str; {'log', 'scale', 'nfft', 'lomb_scargle', 'welch'}

        freq_kwargs : dict
        
            Arguments for frequency vector

        settings : dict
        
            Arguments for the specific spectral method

        verbose : bool
        
            If True, will print warning messages if there is any

        mute_pbar : bool, optional
        
            Whether to mute the progress bar. The default is False.

        Returns
        -------

        scals : MultipleScalograms
        
            A Multiple Scalogram object

        See also
        --------
        
        pyleoclim.utils.wavelet.wwz : wwz function

        pyleoclim.utils.wavelet.cwt : cwt function

        pyleoclim.utils.wavelet.make_freq_vector : Functions to create the frequency vector

        pyleoclim.utils.tsutils.detrend : Detrending function

        pyleoclim.core.series.Series.wavelet : wavelet analysis on single object

        pyleoclim.core.scalograms.MultipleScalogram : Multiple Scalogram object

        References
        ----------

        Torrence, C. and G. P. Compo, 1998: A Practical Guide to Wavelet Analysis. Bull. Amer. Meteor. Soc., 79, 61-78.
        Python routines available at http://paos.colorado.edu/research/wavelets/
        

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            tslist = data.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only contain age and depth
            ms = pyleo.MultipleSeries(tslist)
            wav = ms.wavelet(method='wwz')

        '''
        settings = {} if settings is None else settings.copy()

        scal_list = []
        for s in tqdm(self.series_list, desc='Performing wavelet analysis on individual series', position=0, leave=True, disable=mute_pbar):
            scal_tmp = s.wavelet(method=method, settings=settings, freq_method=freq_method, freq_kwargs=freq_kwargs, verbose=verbose)
            scal_list.append(scal_tmp)

        scals = MultipleScalogram(scalogram_list=scal_list)

        return scals

    def plot(self, figsize=[10, 4],
             marker=None, markersize=None,
             linestyle=None, linewidth=None, colors=None, cmap='tab10', norm=None,
             xlabel=None, ylabel=None, title=None,
             legend=True, plot_kwargs=None, lgd_kwargs=None,
             savefig_settings=None, ax=None, invert_xaxis=False):

        '''Plot multiple timeseries on the same axis

        Parameters
        ----------
        
        figsize : list, optional
        
            Size of the figure. The default is [10, 4].
            
        marker : str, optional
        
            Marker type. The default is None.
            
        markersize : float, optional
        
            Marker size. The default is None.
            
        linestyle : str, optional
        
            Line style. The default is None.
            
        linewidth : float, optional
        
            The width of the line. The default is None.
            
        colors : a list of, or one, Python supported color code (a string of hex code or a tuple of rgba values)
        
            Colors for plotting.
            If None, the plotting will cycle the 'tab10' colormap;
            if only one color is specified, then all curves will be plotted with that single color;
            if a list of colors are specified, then the plotting will cycle that color list.
            
        cmap : str
        
            The colormap to use when "colors" is None.
            
        norm : matplotlib.colors.Normalize
       
            The normalization for the colormap.
            If None, a linear normalization will be used.
            
        xlabel : str, optional
        
            x-axis label. The default is None.
            
        ylabel : str, optional
        
            y-axis label. The default is None.
            
        title : str, optional
        
            Title. The default is None.
            
        legend : bool, optional
        
            Whether the show the legend. The default is True.
            
        plot_kwargs : dict, optional
        
            Plot parameters. The default is None.
            
        lgd_kwargs : dict, optional
        
            Legend parameters. The default is None.
            
        savefig_settings : dictionary, optional
        
            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existing or non-existing path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"} The default is None.
            
        ax : matplotlib.ax, optional
        
            The matplotlib axis onto which to return the figure. The default is None.
            
        invert_xaxis : bool, optional
        
            if True, the x-axis of the plot will be inverted

        Returns
        -------

        fig : matplotlib.figure
        
            the figure object from matplotlib
            See [matplotlib.pyplot.figure](https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.figure.html) for details.

        ax : matplotlib.axis
        
            the axis object from matplotlib
            See [matplotlib.axes](https://matplotlib.org/api/axes_api.html) for details.

        See also
        --------

        pyleoclim.utils.plotting.savefig : Saving figure in Pyleoclim

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            tslist = data.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)

            @savefig ms_basic_plot.png
            fig, ax = ms.plot()
            pyleo.closefig(fig) #Optional close fig after plotting

        '''
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        plot_kwargs = {} if plot_kwargs is None else plot_kwargs.copy()
        lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)

        if ylabel is None:
            consistent_ylabels = True
            time_label, value_label = self.series_list[0].make_labels()
            for s in self.series_list[1:]:
                time_label_tmp, value_label_tmp = s.make_labels()
                if value_label_tmp != value_label:
                    consistent_ylabels = False

            if consistent_ylabels:
                ylabel = value_label
            else:
                ylabel = 'value'

        for idx, s in enumerate(self.series_list):
            if colors is None:
                cmap_obj = plt.get_cmap(cmap)
                if hasattr(cmap_obj, 'colors'):
                    nc = len(cmap_obj.colors)
                else:
                    nc = len(self.series_list)

                if norm is None:
                    norm = mpl.colors.Normalize(vmin=0, vmax=nc-1)

                clr = cmap_obj(norm(idx%nc))
            elif type(colors) is str:
                clr = colors
            elif type(colors) is list:
                nc = len(colors)
                clr = colors[idx%nc]
            else:
                raise TypeError('"colors" should be a list of, or one, Python supported color code (a string of hex code or a tuple of rgba values)')

            ax = s.plot(
                figsize=figsize, marker=marker, markersize=markersize, color=clr, linestyle=linestyle,
                linewidth=linewidth, label=s.label, xlabel=xlabel, ylabel=ylabel, title=title,
                legend=legend, lgd_kwargs=lgd_kwargs, plot_kwargs=plot_kwargs, ax=ax,
            )

        if invert_xaxis:
            ax.invert_xaxis()

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            return fig, ax
        else:
            return ax

    def stackplot(self, figsize=None, savefig_settings=None,  xlim=None, fill_between_alpha=0.2, colors=None, cmap='tab10', norm=None, labels='auto',
                  spine_lw=1.5, grid_lw=0.5, label_x_loc=-0.15, v_shift_factor=3/4, linewidth=1.5, plot_kwargs=None):
        ''' Stack plot of multiple series

        Note that the plotting style is uniquely designed for this one and cannot be properly reset with `pyleoclim.set_style()`.

        Parameters
        ----------

        figsize : list
        
            Size of the figure.
            
        savefig_settings : dictionary
        
            the dictionary of arguments for plt.savefig(); some notes below:
            - "path" must be specified; it can be any existing or non-existing path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"} The default is None.
            
        xlim : list
        
            The x-axis limit.
            
        fill_between_alpha : float
        
            The transparency for the fill_between shades.
            
        colors : a list of, or one, Python supported color code (a string of hex code or a tuple of rgba values)
        
            Colors for plotting.
            If None, the plotting will cycle the 'tab10' colormap;
            if only one color is specified, then all curves will be plotted with that single color;
            if a list of colors are specified, then the plotting will cycle that color list.
            
        cmap : str
        
            The colormap to use when "colors" is None.
            
        norm : matplotlib.colors.Normalize like
        
            The nomorlization for the colormap.
            If None, a linear normalization will be used.
            
        labels: None, 'auto' or list
        
            If None, doesn't add labels to the subplots
            If 'auto', uses the labels passed during the creation of pyleoclim.Series
            If list, pass a list of strings for each labels.
            Default is 'auto'
            
        spine_lw : float
        
            The linewidth for the spines of the axes.
            
        grid_lw : float
        
            The linewidth for the gridlines.
            
        label_x_loc : float
        
            The x location for the label of each curve.
            
        v_shift_factor : float
        
            The factor for the vertical shift of each axis.
            The default value 3/4 means the top of the next axis will be located at 3/4 of the height of the previous one.
            
        linewidth : float
        
            The linewidth for the curves.
            
        plot_kwargs: dict or list of dict
        
            Arguments to further customize the plot from matplotlib.pyplot.plot.

            - Dictionary: Arguments will be applied to all lines in the stackplots
            - List of dictionary: Allows to customize one line at a time.

        Returns
        -------

        fig : matplotlib.figure
            the figure object from matplotlib
            See [matplotlib.pyplot.figure](https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.figure.html) for details.

        ax : matplotlib.axis
            the axis object from matplotlib
            See [matplotlib.axes](https://matplotlib.org/api/axes_api.html) for details.

        See also
        --------

        pyleoclim.utils.plotting.savefig : Saving figure in Pyleoclim

        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            d = pyleo.Lipd(usr_path = url)
            tslist = d.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)
            @savefig mts_stackplot.png
            fig, ax = ms.stackplot()
            pyleo.closefig(fig)

        Let's change the labels on the left

        .. ipython:: python
            :okwarning:
            :okexcept:

            sst = d.to_LipdSeries(number=5)
            d18Osw = d.to_LipdSeries(number=3)
            ms = pyleo.MultipleSeries([sst,d18Osw])

            @savefig mts_stackplot_customlabels.png
            fig, ax = ms.stackplot(labels=['sst','d18Osw'])
            pyleo.closefig(fig) 

        And let's remove them completely

        .. ipython:: python
            :okwarning:
            :okexcept:

            @savefig mts_stackplot_nolabels.png
            fig, ax = ms.stackplot(labels=None)
            pyleo.closefig(fig) #Optional figure close after plotting

        Now, let's add markers to the timeseries.

        .. ipython:: python
            :okwarning:
            :okexcept:

            @savefig mts_stackplot_samemarkers.png
            fig, ax = ms.stackplot(labels=None, plot_kwargs={'marker':'o'})
            pyleo.closefig(fig) #Optional figure close after plotting

        Using different marker types on each series:

        .. ipython:: python
            :okwarning:
            :okexcept:

            @savefig mts_stackplot_differentmarkers.png
            fig, ax = ms.stackplot(labels=None, plot_kwargs=[{'marker':'o'},{'marker':'^'}])
            pyleo.closefig(fig) #Optional figure close after plotting

        '''
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()

        n_ts = len(self.series_list)

        if type(labels)==list:
            if len(labels) != n_ts:
                raise ValueError("The length of the label list should match the number of timeseries to be plotted")

        # Deal with plotting arguments
        if type(plot_kwargs)==dict:
            plot_kwargs = [plot_kwargs]*n_ts

        if plot_kwargs is not None and len(plot_kwargs) != n_ts:
            raise ValueError("When passing a list of dictionaries for kwargs arguments, the number of items should be the same as the number of timeseries")


        fig = plt.figure(figsize=figsize)

        if xlim is None:
            time_min = np.inf
            time_max = -np.inf
            for ts in self.series_list:
                if np.min(ts.time) <= time_min:
                    time_min = np.min(ts.time)
                if np.max(ts.time) >= time_max:
                    time_max = np.max(ts.time)
            xlim = [time_min, time_max]

        ax = {}
        left = 0
        width = 1
        height = 1/n_ts
        bottom = 1
        for idx, ts in enumerate(self.series_list):
            if colors is None:
                cmap_obj = plt.get_cmap(cmap)
                if hasattr(cmap_obj, 'colors'):
                    nc = len(cmap_obj.colors)
                else:
                    nc = len(self.series_list)

                if norm is None:
                    norm = mpl.colors.Normalize(vmin=0, vmax=nc-1)

                clr = cmap_obj(norm(idx%nc))
            elif type(colors) is str:
                clr = colors
            elif type(colors) is list:
                nc = len(colors)
                clr = colors[idx%nc]
            else:
                raise TypeError('"colors" should be a list of, or one, Python supported color code (a string of hex code or a tuple of rgba values)')
            #deal with other plotting arguments
            if plot_kwargs is None:
                p_kwargs = {}
            else:
                p_kwargs = plot_kwargs[idx]

            bottom -= height*v_shift_factor
            ax[idx] = fig.add_axes([left, bottom, width, height])
            ax[idx].plot(ts.time, ts.value, color=clr, lw=linewidth,**p_kwargs)
            ax[idx].patch.set_alpha(0)
            ax[idx].set_xlim(xlim)
            time_label, value_label = ts.make_labels()
            ax[idx].set_ylabel(value_label, weight='bold')

            mu = np.nanmean(ts.value)
            std = np.nanstd(ts.value)
            ylim = [mu-4*std, mu+4*std]
            ax[idx].fill_between(ts.time, ts.value, y2=mu, alpha=fill_between_alpha, color=clr)
            trans = transforms.blended_transform_factory(ax[idx].transAxes, ax[idx].transData)
            if labels == 'auto':
                if ts.label is not None:
                    ax[idx].text(label_x_loc, mu, ts.label, horizontalalignment='right', transform=trans, color=clr, weight='bold')
            elif type(labels) ==list:
                ax[idx].text(label_x_loc, mu, labels[idx], horizontalalignment='right', transform=trans, color=clr, weight='bold')
            elif labels==None:
                pass
            ax[idx].set_ylim(ylim)
            ax[idx].set_yticks(ylim)
            ax[idx].yaxis.set_major_formatter(FormatStrFormatter('%.1f'))
            ax[idx].grid(False)
            if idx % 2 == 0:
                ax[idx].spines['left'].set_visible(True)
                ax[idx].spines['left'].set_linewidth(spine_lw)
                ax[idx].spines['left'].set_color(clr)
                ax[idx].spines['right'].set_visible(False)
                ax[idx].yaxis.set_label_position('left')
                ax[idx].yaxis.tick_left()
            else:
                ax[idx].spines['left'].set_visible(False)
                ax[idx].spines['right'].set_visible(True)
                ax[idx].spines['right'].set_linewidth(spine_lw)
                ax[idx].spines['right'].set_color(clr)
                ax[idx].yaxis.set_label_position('right')
                ax[idx].yaxis.tick_right()

            ax[idx].yaxis.label.set_color(clr)
            ax[idx].tick_params(axis='y', colors=clr)
            ax[idx].spines['top'].set_visible(False)
            ax[idx].spines['bottom'].set_visible(False)
            ax[idx].tick_params(axis='x', which='both', length=0)
            ax[idx].set_xlabel('')
            ax[idx].set_xticklabels([])
            xt = ax[idx].get_xticks()[1:-1]
            for x in xt:
                ax[idx].axvline(x=x, color='lightgray', linewidth=grid_lw, ls='-', zorder=-1)
            ax[idx].axhline(y=mu, color='lightgray', linewidth=grid_lw, ls='-', zorder=-1)

        bottom -= height*(1-v_shift_factor)
        ax[n_ts] = fig.add_axes([left, bottom, width, height])
        ax[n_ts].set_xlabel(time_label)
        ax[n_ts].spines['left'].set_visible(False)
        ax[n_ts].spines['right'].set_visible(False)
        ax[n_ts].spines['bottom'].set_visible(True)
        ax[n_ts].spines['bottom'].set_linewidth(spine_lw)
        ax[n_ts].set_yticks([])
        ax[n_ts].patch.set_alpha(0)
        ax[n_ts].set_xlim(xlim)
        ax[n_ts].grid(False)
        ax[n_ts].tick_params(axis='x', which='both', length=3.5)
        xt = ax[n_ts].get_xticks()[1:-1]
        for x in xt:
            ax[n_ts].axvline(x=x, color='lightgray', linewidth=grid_lw, ls='-', zorder=-1)

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            return fig, ax
        else:
            return ax
        
    def stripes(self, ref_period=None, figsize=None, savefig_settings=None,  
                LIM = 2.8, thickness=1.0, labels='auto',  label_color = 'gray',
                common_time_kwargs=None, xlim=None, font_scale=0.8, x_offset = 0.05):
        '''
        Represents a MultipleSeries object as a quilt of Ed Hawkins' "stripes" patterns
        
        To ensure comparability, constituent series are placed on a common time axis, using
        `MultipleSeries.common_time()`. To ensure consistent scaling, all series are Gaussianized
        prior to plotting. 
    
        Credit: https://showyourstripes.info/,
        Implementation: https://matplotlib.org/matplotblog/posts/warming-stripes/

        Parameters
        ----------

        ref_period : TYPE, optional
            dates of the reference period, in the form "(first, last)".
            The default is None, which will pick the beginning and end of the common time axis.
        
        LIM : float
            scaling factor for color saturation. default is 2.8. 
            The higher the LIM, the more compressed the color range (milder hues)
        
        thickness : float, optional
            vertical thickness of the stripe . The default is 1.0
            
        figsize : list
        
            Size of the figure.
            
        savefig_settings : dictionary
        
            the dictionary of arguments for plt.savefig(); some notes below:

            - 'path' must be specified; it can be any existing or non-existing path,
              with or without a suffix; if the suffix is not given in 'path', it will follow 'format'
            - 'format' can be one of {"pdf", 'eps', 'png', ps'} The default is None.
            
        xlim : list
            The x-axis limit.
            
        x_offset : float
            value controlling the horizontal offset between stripes and labels (default = 0.05)          
            
        labels: None, 'auto' or list
        
            If None, doesn't add labels to the subplots

            If 'auto', uses the labels passed during the creation of pyleoclim.Series

            If list, pass a list of strings for each labels.
            Default is 'auto'
            
        common_time_kwargs : dict
            Optional arguments for common_time()
            
        font_scale : float
            The scale for the font sizes. Default is 0.8.   

        Returns
        -------

        fig : matplotlib.figure
            the figure object from matplotlib
            See [matplotlib.pyplot.figure](https://matplotlib.org/stable/api/figure_api.html) for details.

        ax : matplotlib.axis
            the axis object from matplotlib
            See [matplotlib.axes](https://matplotlib.org/stable/api/axes_api.html) for details.
            
        See also
        --------
        
        pyleoclim.core.multipleseries.MultipleSeries.common_time : aligns the time axes of a MultipleSeries object

        pyleoclim.utils.plotting.savefig : saving a figure in Pyleoclim
        
        pyleoclim.core.series.Series.stripes : stripes representation in Pyleoclim   
        
        pyleoclim.utils.tsutils.gaussianize : mapping to a standard Normal distribution
            
        Examples
        --------

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            d = pyleo.Lipd(usr_path = url)
            tslist = d.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)
            @savefig md76_stripes.png
            fig, ax = ms.stripes()
            pyleo.closefig(fig)
             
        The default style has rather thick bands, intense colors, and too many stripes.
        The first issue can be solved by passing a figsize tuple; the second by increasing the LIM parameter; 
        the third by passing a step of 0.5 (500y) to common_time(). Finally, the 
        labels are too close to the edge of the plot, which can be adjusted with x_offset, like so:  

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            d = pyleo.Lipd(usr_path = url)
            tslist = d.to_LipdSeriesList()
            tslist = tslist[2:] # drop the first two series which only concerns age and depth
            ms = pyleo.MultipleSeries(tslist)
            @savefig md76_stripes2.png
            fig, ax = ms.stripes(common_time_kwargs={'step': 0.5}, x_offset = 200, 
                                 LIM=4, figsize=[8,3])
            pyleo.closefig(fig)     
            
        '''
        current_style = deepcopy(mpl.rcParams)
        plotting.set_style('journal', font_scale=font_scale)
        savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
        common_time_kwargs = {} if common_time_kwargs is None else common_time_kwargs.copy()

        # put on common timescale
        msc = self.common_time(**common_time_kwargs)
        
        ts0 = msc.series_list[0]
        time = ts0.time
        # generate default axis labels
        time_label, _ = ts0.make_labels()
           
        if ref_period is None:
            ref_period = [time.min(), time.max()]
        
        n_ts = len(msc.series_list)
        last = n_ts-1

        if n_ts < 2:
            raise ValueError("There is only one series in this object. Please use the Series class instead")

        if type(labels)==list:
            if len(labels) != n_ts:
                raise ValueError("The length of the label list should match the number of timeseries to be plotted")

        fig, axs = plt.subplots(n_ts, 1, sharex=True, figsize=figsize, layout = 'tight')
        ax = axs.flatten()

        if xlim is None:
            xlim = [time.min(), time.max()]

        for idx in range(n_ts-1):  # loop over series
            ts = msc.series_list[idx].gaussianize()
            ts.stripes(ref_period, LIM = LIM, label_color = label_color,
                       ax=ax[idx], x_offset=x_offset) 
            
        # handle bottom plot
        ts = msc.series_list[last].gaussianize()
        ts.stripes(ref_period, LIM = LIM, label_color = label_color, 
                   ax=ax[last], x_offset=x_offset, show_xaxis=True) 
        ax[last].set_xlabel(time_label)
        ax[last].set_xlim(xlim)

        if 'fig' in locals():
            if 'path' in savefig_settings:
                plotting.savefig(fig, settings=savefig_settings)
            mpl.rcParams.update(current_style)
            return fig, ax
        else:
            # reset the plotting style
            mpl.rcParams.update(current_style)
            return ax



''' Tests for pyleoclim.core.ui.CorrEns class

Naming rules:
1. class: Test{filename}{Class}{method} with appropriate camel case
2. function: test_{method}_t{test_id}

Notes on how to test:
0. Make sure [pytest](https://docs.pytest.org) has been installed: `pip install pytest`
1. execute `pytest {directory_path}` in terminal to perform all tests in all testing files inside the specified directory
2. execute `pytest {file_path}` in terminal to perform all tests in the specified file
3. execute `pytest {file_path}::{TestClass}::{test_method}` in terminal to perform a specific test class/method inside the specified file
4. after `pip install pytest-xdist`, one may execute "pytest -n 4" to test in parallel with number of workers specified by `-n`
5. for more details, see https://docs.pytest.org/en/stable/usage.html
'''
import numpy as np
import matplotlib.pyplot as plt

import pytest

import pyleoclim as pyleo

def gen_ts(model,nt,alpha):
    t,v = pyleo.utils.gen_ts(model=model,nt=nt,alpha=alpha)
    ts=pyleo.Series(t,v)
    return ts

# Tests below
class TestUiCorrEns():
    def test_plot_t0(self):
        ''' Test CorrEns.plot() for multiple plots
            (qui peut le plus peut le moins)
        '''
        nn = 20 # number of noise realizations
        nt = 200
        
        signal = gen_ts(model='colored_noise',nt=nt,alpha=1.0).standardize() 
        noise = np.random.randn(nt,nn)

        list1 = []
        list2 = []
        nhlf = int(nn/2)
        for idx in range(nhlf):  
            ts1 = pyleo.Series(time=signal.time, value=signal.value+noise[:,idx])  
            list1.append(ts1)
            ts2 = pyleo.Series(time=signal.time, value=signal.value+noise[:,idx+nhlf])
            list2.append(ts2)

        ts_ens1 = pyleo.EnsembleSeries(list1)
        ts_ens2 = pyleo.EnsembleSeries(list2)

        fig, axs = plt.subplots(1,2)
        corr1 = ts_ens1.correlation(signal,settings={'nsim':100})
        corr1.plot(ax=axs[0])
        corr2 = ts_ens2.correlation(signal,settings={'nsim':100})
        corr2.plot(ax=axs[1])
        pyleo.closefig(fig)
        
        




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Plotting utilities, leveraging Matplotlib.
"""

__all__ = ['set_style','closefig', 'savefig']


import matplotlib.pyplot as plt
import pathlib
import matplotlib as mpl
import numpy as np
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection
from matplotlib.colors import ListedColormap


def scatter_xy(x,y,c=None, figsize=None, xlabel=None, ylabel=None, title=None, 
            xlim=None, ylim=None, savefig_settings=None, ax=None,
            legend=True, plot_kwargs=None, lgd_kwargs=None):
    """
    Make scatter plot. 

    Parameters
    ----------
    x : numpy.array
        x value
    y : numpy.array
        y value
    c : TYPE, optional
        DESCRIPTION. The default is None.
    figsize : list, optional
        A list of two integers indicating the dimension of the figure. The default is None.
    xlabel : str, optional
        x-axis label. The default is None.
    ylabel : str, optional
        y-axis label. The default is None.
    title : str, optional
        Title for the plot. The default is None.
    xlim : list, optional
        Limits for the x-axis. The default is None.
    ylim : list, optional
        Limits for the y-axis. The default is None.
    savefig_settings : dict, optional
        the dictionary of arguments for plt.savefig(); some notes below:
        - "path" must be specified; it can be any existing or non-existing path,
          with or without a suffix; if the suffix is not given in "path", it will follow "format"
        - "format" can be one of {"pdf", "eps", "png", "ps"}
       The default is None.
    ax : pyplot.axis, optional
        The axis object. The default is None.
    legend : bool, optional
        Whether to include a legend. The default is True.
    plot_kwargs : dict, optional
        the keyword arguments for ax.plot(). The default is None.
    lgd_kwargs : dict, optional
        the keyword arguments for ax.legend(). The default is None.

    Returns
    -------
    ax : the pyplot.axis object

    """
    savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
    plot_kwargs = {} if plot_kwargs is None else plot_kwargs.copy()
    lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()
    
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    ax.scatter(x, y, c=c, **plot_kwargs)

    if xlabel is not None:
        ax.set_xlabel(xlabel)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if title is not None:
        ax.set_title(title)

    if xlim is not None:
        ax.set_xlim(xlim)

    if ylim is not None:
        ax.set_ylim(ylim)

    if legend:
        ax.legend(**lgd_kwargs)
    else:
        ax.legend().remove()

    if 'fig' in locals():
        if 'path' in savefig_settings:
            savefig(fig, settings=savefig_settings)
        return fig, ax
    else:
        return ax


def plot_scatter_xy(x1,y1,x2,y2, figsize=None, xlabel=None,
                    ylabel=None, title=None, xlim=None, ylim=None,
                    savefig_settings=None, ax=None, legend=True, 
                    plot_kwargs=None, lgd_kwargs=None):
    
    ''' Plot a scatter on top of a line plot.
    
    Parameters
    ----------
    
    x1 : array
      x axis of timeseries1 - plotted as a line
    y1 : array
     values of timeseries1 - plotted as a line
    x2 : array
        x axis of scatter points
    y2 : array
        y of scatter points
    figsize : list
        a list of two integers indicating the figure size
    xlabel : str
        label for x-axis
    ylabel : str
        label for y-axis
    title : str
        the title for the figure
    xlim : list
        set the limits of the x axis
    ylim : list
        set the limits of the y axis
    ax : pyplot.axis
        the pyplot.axis object
    legend : bool
        plot legend or not
    lgd_kwargs : dict
        the keyword arguments for ax.legend()
    plot_kwargs : dict
        the keyword arguments for ax.plot()
    savefig_settings : dict
        the dictionary of arguments for plt.savefig(); some notes below:
        - "path" must be specified; it can be any existing or non-existing path,
          with or without a suffix; if the suffix is not given in "path", it will follow "format"
        - "format" can be one of {"pdf", "eps", "png", "ps"}

    Returns
    -------
    
    ax : the pyplot.axis object
    
    See also
    -------- 
    
    pyleoclim.utils.plotting.set_style : set different styles for the figures. Should be set before invoking the plotting functions
    pyleoclim.utils.plotting.savefig : save figures
    
    '''
    # handle dict defaults
    savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
    plot_kwargs = {} if plot_kwargs is None else plot_kwargs.copy()
    lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    ax.plot(x1, y1, **plot_kwargs, color='green')
    ax.scatter(x2, y2, color='red')

    if xlabel is not None:
        ax.set_xlabel(xlabel)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if title is not None:
        ax.set_title(title)

    if xlim is not None:
        ax.set_xlim(xlim)

    if ylim is not None:
        ax.set_ylim(ylim)

    if legend:
        ax.legend(**lgd_kwargs)
    else:
        ax.legend().remove()

    if 'fig' in locals():
        if 'path' in savefig_settings:
            savefig(fig, settings=savefig_settings)
        return fig, ax
    else:
        return ax


def plot_xy(x, y, figsize=None, xlabel=None, ylabel=None, title=None, 
            xlim=None, ylim=None,savefig_settings=None, ax=None,
            legend=True, plot_kwargs=None, lgd_kwargs=None, 
            invert_xaxis=False, invert_yaxis=False):
    ''' Plot a timeseries
    
    Parameters
    ----------
    x : array
        The time axis for the timeseries
    y : array
        The values of the timeseries
    figsize : list
        a list of two integers indicating the figure size
    xlabel : str
        label for x-axis
    ylabel : str
        label for y-axis
    title : str
        the title for the figure
    xlim : list
        set the limits of the x axis
    ylim : list
        set the limits of the y axis
    ax : pyplot.axis
        the pyplot.axis object
    legend : bool
        plot legend or not
    lgd_kwargs : dict
        the keyword arguments for ax.legend()
    plot_kwargs : dict
        the keyword arguments for ax.plot()
    mute : bool
        if True, the plot will not show;
        recommend to turn on when more modifications are going to be made on ax
         (going to be deprecated)
    savefig_settings : dict
        the dictionary of arguments for plt.savefig(); some notes below:
        - "path" must be specified; it can be any existing or non-existing path,
          with or without a suffix; if the suffix is not given in "path", it will follow "format"
        - "format" can be one of {"pdf", "eps", "png", "ps"}
    invert_xaxis : bool, optional
        if True, the x-axis of the plot will be inverted
    invert_yaxis : bool, optional
        if True, the y-axis of the plot will be inverted
        
    Returns
    -------

    ax : the pyplot.axis object

    See Also
    --------
    
    pyleoclim.utils.plotting.set_style : set different styles for the figures. Should be set before invoking the plotting functions
    pyleoclim.utils.plotting.savefig : save figures        
    '''
    # handle dict defaults
    savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
    plot_kwargs = {} if plot_kwargs is None else plot_kwargs.copy()
    lgd_kwargs = {} if lgd_kwargs is None else lgd_kwargs.copy()

    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    ax.plot(x, y, **plot_kwargs)

    if xlabel is not None:
        ax.set_xlabel(xlabel)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if title is not None:
        ax.set_title(title)
        # TODO replace with ax.set_title(title, fontweight='bold') when all relevant plots use plot_xy

    if xlim is not None:
        ax.set_xlim(xlim)

    if ylim is not None:
        ax.set_ylim(ylim)

    if legend:
        ax.legend(**lgd_kwargs)
    else:
        ax.legend().remove()

    if invert_xaxis:
        ax.invert_xaxis()
        
    if invert_yaxis:
        ax.invert_yaxis()

    if 'fig' in locals():
        if 'path' in savefig_settings:
            savefig(fig, settings=savefig_settings)
        return fig, ax
    else:
        return ax
    
def stripes_xy(x, y, ref_period, thickness = 1.0, LIM = 0.75, figsize=None, xlabel=None, 
               ylabel=None, title=None, xlim=None, savefig_settings=None, ax=None, 
               x_offset = 0.05, label_size = None, show_xaxis = False,
               invert_xaxis=False, top_label = None, bottom_label = None, label_color = None): 
    '''
    Represent y as an Ed Hawkins "warming stripes" pattern, as a function of x
    
    Credit: https://matplotlib.org/matplotblog/posts/warming-stripes/
    
    Parameters
    ----------
    x : array
        Independent variable
    y : array
        Dependent variable
    ref_period : 2-tuple or 2-vector
        indices of the reference period, in the form "(first, last)"
    thickness : float, optional
        vertical thickness of the stripe . The default is 1.0     
    LIM : float, optional 
        size of the y-value range (default: 0.7) 
    figsize : list
        a list of two integers indicating the figure size
    top_label : str
        the "title" label for the stripe
    bottom_label : str
        the "ylabel" explaining which variable is being plotted
    label_size : int
        size of the text in labels (in points). Default is the Matplotlib 'axes.labelsize'] rcParams
    xlim : list
        set the limits of the x axis
    x_offset : float
        value controlling the horizontal offset between stripes and labels (default = 0.05)
    show_xaxis : bool
        flag indicating whether or not the x-axis should be shown (default = False)
    ax : pyplot.axis
        the pyplot.axis object
    savefig_settings : dict
        the dictionary of arguments for plt.savefig(); some notes below:
        - "path" must be specified; it can be any existing or non-existing path,
          with or without a suffix; if the suffix is not given in "path", it will follow "format"
        - "format" can be one of {"pdf", "eps", "png", "ps"}
    invert_xaxis : bool, optional
        if True, the x-axis of the plot will be inverted
   
    Returns
    -------
    ax, or fig, ax (if no axes were provided)

    '''
    # handle dict defaults
    savefig_settings = {} if savefig_settings is None else savefig_settings.copy()
    
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
      
    if label_size is None:
        label_size = mpl.rcParams['axes.labelsize']
      
    if thickness is None:
        thickness = 5*label_size
        
    ax.get_yaxis().set_visible(False) # remove parasitic lines and labels
    ax.get_xaxis().set_visible(show_xaxis) # remove parasitic lines and labels
    ax.spines[:].set_visible(False)
    
    dx = np.diff(x).mean()
    xmin = x.min()
    xmax = x.max() + dx # inclusive
    
    # Reference period for the center of the color scale    
    reference = y[ref_period[0]:ref_period[1]].mean()
    
    # colormap: the 8 more saturated colors from the 9 blues / 9 reds
    cmap = ListedColormap([
        '#08306b', '#08519c', '#2171b5', '#4292c6',
        '#6baed6', '#9ecae1', '#c6dbef', '#deebf7',
        '#fee0d2', '#fcbba1', '#fc9272', '#fb6a4a',
        '#ef3b2c', '#cb181d', '#a50f15', '#67000d',
    ])
    
    col = PatchCollection([
        Rectangle((yl, 0), 1, 1)
        for yl in range(int(xmin), int(xmax))
        ])

    # set data, colormap and color limits
    col.set_array(y)
    col.set_cmap(cmap)
    col.set_clim(reference - LIM, reference + LIM)
    ax.add_collection(col)
    # adjust axes
    ax.set_ylim(0, thickness)
    ax.set_xlim(xmin, xmax);
    
    # add label to the right 
    #offset = y_offsets[column] / 72
    #trans = mtransforms.ScaledTranslation(0, offset, fig.dpi_scale_trans)
    #trans = ax.transData #+ trans
    
    ypos = 0.4*thickness
    ax.text(xmax+dx+x_offset, 0.6*thickness, top_label, color=label_color, 
            fontsize=label_size, fontweight = 'bold')
    ax.text(xmax+dx+x_offset, 0.2*thickness, bottom_label, color=label_color,
            fontsize=label_size)

    if xlabel is not None:
        ax.set_xlabel(xlabel)

    if ylabel is not None:
        ax.set_ylabel(ylabel)

    if title is not None:
        ax.set_title(title)

    if xlim is not None:
        ax.set_xlim(xlim)

    if invert_xaxis:
        ax.invert_xaxis()       

    if 'fig' in locals():
        fig.tight_layout()
        if 'path' in savefig_settings:
            savefig(fig, settings=savefig_settings)
        return fig, ax
    else:
        return ax

def closefig(fig=None):
    '''Close the figure

    Parameters
    ----------
    fig : matplotlib.pyplot.figure
        The matplotlib figure object

    '''
    if fig is not None:
        plt.close(fig)
    else:
        plt.close()

def savefig(fig, path=None, dpi=300, settings={}, verbose=True):
    ''' Save a figure to a path

    Parameters
    ----------
    fig : matplotlib.pyplot.figure
        the figure to save
    path : str
        the path to save the figure, can be ignored and specify in "settings" instead
    dpi : int
        resolution in dot (pixels) per inch. Default: 300. 
    settings : dict
        the dictionary of arguments for plt.savefig(); some notes below:
        - "path" must be specified in settings if not assigned with the keyword argument;
          it can be any existing or non-existing path, with or without a suffix;
          if the suffix is not given in "path", it will follow "format"
        - "format" can be one of {"pdf", "eps", "png", "ps"}
    verbose : bool, {True,False}
        If True, print the path of the saved file.
        
    '''
    if path is None and 'path' not in settings:
        raise ValueError('"path" must be specified, either with the keyword argument or be specified in `settings`!')

    savefig_args = {'bbox_inches': 'tight', 'path': path, 'dpi': dpi}
    savefig_args.update(settings)

    path = pathlib.Path(savefig_args['path'])
    savefig_args.pop('path')

    dirpath = path.parent
    if not dirpath.exists():
        dirpath.mkdir(parents=True, exist_ok=True)
        if verbose:
            print(f'Directory created at: "{dirpath}"')

    path_str = str(path)
    if path.suffix not in ['.eps', '.pdf', '.png', '.ps']:
        path = pathlib.Path(f'{path_str}.pdf')

    fig.savefig(path_str, **savefig_args)
    plt.close(fig)

    if verbose:
        print(f'Figure saved at: "{str(path)}"')


def set_style(style='journal', font_scale=1.0, dpi=300):
    ''' Modify the visualization style
    
    This function is inspired by [Seaborn](https://github.com/mwaskom/seaborn).
   
    
    Parameters
    ----------
    
    style : {journal,web,matplotlib,_spines, _nospines,_grid,_nogrid}
        set the styles for the figure:
            - journal (default): fonts appropriate for paper
            - web: web-like font (e.g. ggplot)
            - matplotlib: the original matplotlib style
            In addition, the following options are available:
            - _spines/_nospines: allow to show/hide spines
            - _grid/_nogrid: allow to show gridlines (default: _grid)
    
    font_scale : float
        Default is 1. Corresponding to 12 Font Size. 
    
    '''
    font_dict = {
        'font.size': 10,
        'axes.labelsize': 11,
        'axes.titlesize': 12,
        'xtick.labelsize': 10,
        'ytick.labelsize': 10,
        'legend.fontsize': 9,
    }

    style_dict = {}
    if 'journal' in style:
        style_dict.update({
            'axes.axisbelow': True,
            'axes.facecolor': 'white',
            'axes.edgecolor': 'black',
            'axes.grid': True,
            'grid.color': 'lightgrey',
            'grid.linestyle': '--',
            'xtick.direction': 'out',
            'ytick.direction': 'out',
            'font.sans-serif': ['Arial', 'DejaVu Sans', 'Liberation Sans', 'Bitstream Vera Sans', 'sans-serif'],

            'axes.spines.left': True,
            'axes.spines.bottom': True,
            'axes.spines.right': False,
            'axes.spines.top': False,

            'legend.frameon': False,

            'axes.linewidth': 1,
            'grid.linewidth': 1,
            'lines.linewidth': 2,
            'lines.markersize': 6,
            'patch.linewidth': 1,

            'xtick.major.width': 1.25,
            'ytick.major.width': 1.25,
            'xtick.minor.width': 0,
            'ytick.minor.width': 0,
        })
    elif 'web' in style:
        style_dict.update({
            'figure.facecolor': 'white',

            'axes.axisbelow': True,
            'axes.facecolor': 'whitesmoke',
            'axes.edgecolor': 'lightgrey',
            'axes.grid': True,
            'grid.color': 'white',
            'grid.linestyle': '-',
            'xtick.direction': 'out',
            'ytick.direction': 'out',

            'text.color': 'grey',
            'axes.labelcolor': 'grey',
            'xtick.color': 'grey',
            'ytick.color': 'grey',

            'font.sans-serif': ['Arial', 'DejaVu Sans', 'Liberation Sans', 'Bitstream Vera Sans', 'sans-serif'],

            'axes.spines.left': False,
            'axes.spines.bottom': False,
            'axes.spines.right': False,
            'axes.spines.top': False,

            'legend.frameon': False,

            'axes.linewidth': 1,
            'grid.linewidth': 1,
            'lines.linewidth': 2,
            'lines.markersize': 6,
            'patch.linewidth': 1,

            'xtick.major.width': 1.25,
            'ytick.major.width': 1.25,
            'xtick.minor.width': 0,
            'ytick.minor.width': 0,
        })
    else:
        raise ValueError(f'Style [{style}] not availabel!')

    if '_spines' in style:
        style_dict.update({
            'axes.spines.left': True,
            'axes.spines.bottom': True,
            'axes.spines.right': True,
            'axes.spines.top': True,
        })
    elif '_nospines' in style:
        style_dict.update({
            'axes.spines.left': False,
            'axes.spines.bottom': False,
            'axes.spines.right': False,
            'axes.spines.top': False,
        })

    if '_grid' in style:
        style_dict.update({
            'axes.grid': True,
        })
    elif '_nogrid' in style:
        style_dict.update({
            'axes.grid': False,
        })

    figure_dict = {
        'savefig.dpi': dpi,
    }

    # modify font size based on font scale
    font_dict.update({k: v * font_scale for k, v in font_dict.items()})

    for d in [style_dict, font_dict, figure_dict]:
        mpl.rcParams.update(d)













"""
This allows to manipulate LiPD objects and take advantage of the metadata information for specific functionalities. Lipd objects are needed to create LipdSeries objects, which carry most of the timeseries functionalities.
"""

from ..utils import mapping, lipdutils

from ..core.lipdseries import LipdSeries
from copy import deepcopy
import warnings
import os

import lipd as lpd

class Lipd:
    '''The Lipd class allows to create a Lipd object from Lipd files. 
    This allows to manipulate LiPD objects and take advantage of the metadata information 
    for specific functionalities. Lipd objects are needed to create LipdSeries objects, 
    which carry most of the timeseries functionalities.


    Parameters
    ----------

    usr_path : str
    
        Path to the Lipd file(s). Can be URL (LiPD utilities only support loading one file at a time from a URL).
        If it's a URL, it must start with "http", "https", or "ftp".

    lidp_dict : dict
    
        LiPD files already loaded into Python through the LiPD utilities

    validate : bool
    
        Validate the LiPD files upon loading. Note that for a large library (>300files) this can take up to half an hour.

    remove : bool
    
        If validate is True and remove is True, ignores non-valid LiPD files. Note that loading unvalidated Lipd files may result in errors for some functionalities but not all.

    TODO
    ----

    Support querying the LinkedEarth platform
    
    References
    ----------
    
    McKay, N. P., & Emile-Geay, J. (2016). Technical Note: The Linked Paleo Data framework – a common tongue for paleoclimatology. Climate of the Past, 12, 1093-1100. 

    Examples
    --------

    .. ipython:: python
        :okwarning:
        :okexcept:

        import pyleoclim as pyleo
        url='http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
        d=pyleo.Lipd(usr_path=url)
    '''

    def __init__(self, usr_path=None, lipd_dict=None, validate=False, remove=False):
        self.plot_default = {'ice-other': ['#FFD600','h'],
                'ice/rock': ['#FFD600', 'h'],
                'coral': ['#FF8B00','o'],
                'documents':['k','p'],
                'glacierice':['#86CDFA', 'd'],
                'hybrid': ['#00BEFF','*'],
                'lakesediment': ['#4169E0','s'],
                'marinesediment': ['#8A4513', 's'],
                'sclerosponge' : ['r','o'],
                'speleothem' : ['#FF1492','d'],
                'wood' : ['#32CC32','^'],
                'molluskshells' : ['#FFD600','h'],
                'peat' : ['#2F4F4F','*'],
                'midden' : ['#824E2B','o'],
                'other':['k','o']}

        if validate==False and remove==True:
            print('Removal of unvalidated LiPD files require validation')
            validate=True

        #prepare the dictionaries for all possible scenarios
        if usr_path!=None:
            # since readLipd() takes only absolute path and it will change the current working directory (CWD) without turning back,
            # we need to record CWD manually and turn back after the data loading is finished
            cwd = os.getcwd()
            if usr_path[:4] == 'http' or usr_path[:3] == 'ftp':
                # URL
                D_path = lpd.readLipd(usr_path)
            else:
                # local path
                abs_path = os.path.abspath(usr_path)
                D_path = lpd.readLipd(abs_path)

            os.chdir(cwd)

            #make sure that it's more than one
            if 'archiveType' in D_path.keys():
                D_path={D_path['dataSetName']:D_path}
            if validate==True:
                cwd = os.getcwd()
                res=lpd.validate(D_path,detailed=False)
                os.chdir(cwd)
                if remove == True:
                    for item in res:
                        if item['status'] == 'FAIL':
                           c=item['feedback']['errMsgs']
                           check = []
                           for i in c:
                               if i.startswith('Mismatched columns'):
                                   check.append(1)
                               else: check.append(0)
                           if 0 in check:
                               del D_path[item['filename'].strip('.lpd')]
        else:
            D_path={}
        if lipd_dict!=None:
            D_dict=lipd_dict
            if 'archiveType' in D_dict.keys():
                D_dict={D_dict['dataSetName']:D_dict}
            if validate==True:
                cwd = os.getcwd()
                res=lpd.validate(D_dict,detailed=False)
                os.chdir(cwd)
                if remove == True:
                    for item in res:
                        if item['status'] == 'FAIL':
                           c=item['feedback']['errMsgs']
                           check = []
                           for i in c:
                               if i.startswith('Mismatched columns'):
                                   check.append(1)
                               else: check.append(0)
                           if 0 in check:
                               del D_dict[item['filename'].strip('.lpd')]
        else:
            D_dict={}

        # raise an error if empty
        if not bool(D_dict) and not bool(D_path) == True:
            raise ValueError('No valid files; try without validation.')
        #assemble
        self.lipd={}
        self.lipd.update(D_path)
        self.lipd.update(D_dict)

    def __repr__(self):
        return str(self.__dict__)

    def copy(self):
        '''Copy the object
        '''
        return deepcopy(self)

    def to_tso(self, mode='paleo'):
        '''Extracts all the variables to a list of LiPD timeseries objects
        
        In LiPD, timeseries objects are flatten dictionaries that contain the values for the time and variable axes as well as relevant metadata. 

        Parameters
        ----------
        
        mode : {'paleo','chron'}
        
            Whether to extract the timeseries information from the paleo tables or chron tables

        Returns
        -------
        
        ts_list : list
        
            List of LiPD timeseries objects

        
        References
        ----------
    
        McKay, N. P., & Emile-Geay, J. (2016). Technical Note: The Linked Paleo Data framework – a common tongue for paleoclimatology. Climate of the Past, 12, 1093-1100. 
        '''
        cwd = os.getcwd()
        ts_list=lpd.extractTs(self.__dict__['lipd'], mode=mode)
        os.chdir(cwd)
        return ts_list

    def extract(self,dataSetName):
        '''
        Parameters
        ----------
        
        dataSetName : str
        
            Extract a particular dataset

        Returns
        -------
        
        new : Lipd
        
            A new object corresponding to a particular dataset

        '''
        new = self.copy()
        try:
            dict_out=self.__dict__['lipd'][dataSetName]
            new.lipd=dict_out
        except:
            pass

        return new

    def to_LipdSeriesList(self, mode='paleo'):
        '''Extracts all LiPD timeseries objects to a list of LipdSeries objects
        
        In LiPD, timeseries objects are flatten dictionaries that contain the values for the time and variable axes as well as relevant metadata. 

        Parameters
        ----------

        mode : {'paleo','chron'}
        
            Whether to extract the timeseries information from the paleo tables or chron tables

        Returns
        -------
        res : list
        
            A list of LiPDSeries objects
        
        References
        ----------
    
        McKay, N. P., & Emile-Geay, J. (2016). Technical Note: The Linked Paleo Data framework – a common tongue for paleoclimatology. Climate of the Past, 12, 1093-1100. 

        See also
        --------
        
        pyleoclim.core.lipdseries.LipdSeries : a LipdSeries object

        '''
        cwd = os.getcwd()
        ts_list=lpd.extractTs(self.__dict__['lipd'], mode=mode)
        os.chdir(cwd)

        res=[]

        for idx, item in enumerate(ts_list):
            try:
                res.append(LipdSeries(item))
            except:
                if mode == 'paleo':
                    txt = 'The timeseries from ' + str(idx) + ': ' +\
                            item['dataSetName'] + ': ' + \
                            item['paleoData_variableName'] + \
                            ' could not be coerced into a LipdSeries object, passing'
                else:
                    txt = 'The timeseries from ' + str(idx) + ': ' +\
                            item['dataSetName'] + ': ' + \
                            item['chronData_variableName'] + \
                            ' could not be coerced into a LipdSeries object, passing'
                warnings.warn(txt)
                pass

        return res

    def to_LipdSeries(self, number = None, mode = 'paleo'):
        '''Extracts one timeseries from the Lipd object

        In LiPD, timeseries objects are flatten dictionaries that contain the values for the time and variable axes as well as relevant metadata. 
        Note that this function may require user interaction if the number of the column in the file is unknown. The numbers are fixed so automating the code is as simple as retaining a series of numbers when reopening the files. 

        Parameters
        ----------

        number : int
        
            the number of the timeseries object

        mode : str; {'paleo','chron'}
        
            whether to extract the paleo or chron series.

        Returns
        -------
        
        ts : pyleoclim.LipdSeries
        
            A LipdSeries object

        See also
        --------
        
        pyleoclim.core.lipdseries.LipdSeries : LipdSeries object

        '''
        cwd = os.getcwd()
        ts_list = lpd.extractTs(self.__dict__['lipd'], mode=mode)
        os.chdir(cwd)
        if number is None:
            ts = LipdSeries(ts_list)
        else:
            try:
                number = int(number)
            except:
                raise TypeError('Number needs to be an integer or should be coerced into an integer.')
            ts = LipdSeries(ts_list[number])
        return ts



    def mapAllArchive(self, projection = 'Robinson', proj_default = True,
           background = True,borders = False, rivers = False, lakes = False,
           figsize = None, ax = None, marker=None, color=None,
           markersize = None, scatter_kwargs=None,
           legend=True, lgd_kwargs=None, savefig_settings=None):

        '''Map all the records contained in the LiPD object by the type of archive
        
        Note that the map is fully cusomizable by using the optional parameters. 

        Parameters
        ----------
        
        projection : str, optional
        
            The projection to use. The default is 'Robinson'.
            
        proj_default : bool, optional
        
            Wether to use the Pyleoclim defaults for each projection type. The default is True.
            
        background : bool, optional
        
            Wether to use a backgound. The default is True.
            
        borders : bool, optional
        
            Draw borders. The default is False.
            
        rivers : bool, optional
        
            Draw rivers. The default is False.
            
        lakes : bool, optional
        
            Draw lakes. The default is False.
            
        figsize : list, optional
        
            The size of the figure. The default is None.
            
        ax : matplotlib.ax, optional
        
            The matplotlib axis onto which to return the map. The default is None.
            
        marker : str, optional
        
            The marker type for each archive. The default is None, which uses a pre-defined palette in Pyleoclim.  
            To see the default option, run Lipd.plot_default where Lipd is the name of the object. 
            
        color : str, optional
        
            Color for each acrhive. The default is None. The default is None, which uses a pre-defined palette in Pyleoclim.  
            To see the default option, run Lipd.plot_default where Lipd is the name of the object. 
            
        markersize : float, optional
        
            Size of the marker. The default is None. 
            
        scatter_kwargs : dict, optional
        
            Parameters for the scatter plot. The default is None.
            
        legend : bool; {True,False}, optional
        
            Whether to plot the legend. The default is True.
            
        lgd_kwargs : dict, optional
        
            Arguments for the legend. The default is None.
            
        savefig_settings : dictionary, optional
        
            The dictionary of arguments for plt.savefig(); some notes below:  
            - "path" must be specified; it can be any existing or non-existing path,
              with or without a suffix; if the suffix is not given in "path", it will follow "format"
            - "format" can be one of {"pdf", "eps", "png", "ps"}. 
            The default is None.
        
        Returns
        -------
        
        res : tuple or fig
        
            The figure and axis if asked. 

        See also
        --------

        pyleoclim.utils.mapping.map : Underlying mapping function for Pyleoclim

        Examples
        --------

        For speed, we are only using one LiPD file. But these functions can load and map multiple.

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            @savefig mapallarchive.png
            fig, ax = data.mapAllArchive()
            pyleo.closefig(fig)

        Change the markersize

        .. ipython:: python
            :okwarning:
            :okexcept:

            import pyleoclim as pyleo
            url = 'http://wiki.linked.earth/wiki/index.php/Special:WTLiPD?op=export&lipdid=MD982176.Stott.2004'
            data = pyleo.Lipd(usr_path = url)
            @savefig mapallarchive_marker.png
            fig, ax = data.mapAllArchive(markersize=100)
            pyleo.closefig(fig)


        '''
        scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs.copy()


        #get the information from the LiPD dict
        lat=[]
        lon=[]
        archiveType=[]

        for idx, key in enumerate(self.lipd):
            d = self.lipd[key]
            lat.append(d['geo']['geometry']['coordinates'][1])
            lon.append(d['geo']['geometry']['coordinates'][0])
            if 'archiveType' in d.keys():
                archiveType.append(lipdutils.LipdToOntology(d['archiveType']).lower().replace(" ",""))
            else:
                archiveType.append('other')

        # make sure criteria is in the plot_default list
        for idx,val in enumerate(archiveType):
            if val not in self.plot_default.keys():
                archiveType[idx] = 'other'

        if markersize is not None:
            scatter_kwargs.update({'s': markersize})

        if marker==None:
            marker=[]
            for item in archiveType:
                marker.append(self.plot_default[item][1])

        if color==None:
            color=[]
            for item in archiveType:
                color.append(self.plot_default[item][0])

        res = mapping.map(lat=lat, lon=lon, criteria=archiveType,
                              marker=marker, color =color,
                              projection = projection, proj_default = proj_default,
                              background = background,borders = borders,
                              rivers = rivers, lakes = lakes,
                              figsize = figsize, ax = ax,
                              scatter_kwargs=scatter_kwargs, legend=legend,
                              lgd_kwargs=lgd_kwargs,savefig_settings=savefig_settings)

        return res


''' Module for timeseries modeling
'''

import numpy as np
# new for statsmodels v0.12
from statsmodels.tsa.arima_process import arma_generate_sample
from statsmodels.tsa.arima.model import ARIMA

from .tsbase import (
    is_evenly_spaced
)
#from .tsutils import preprocess   # no longer used here
from scipy import optimize

__all__ = [
    'ar1_sim',
    'ar1_fit',
    'colored_noise',
    'colored_noise_2regimes',
    'gen_ar1_evenly',
    'gen_ts'
]

def ar1_model(t, tau, output_sigma=1):
    ''' Simulate AR(1) process with REDFIT
    
    Simulate a (possibly irregularly-sampled) AR(1) process with given decay constant tau, à la REDFIT.

    Parameters
    ----------

    t :  array
        Time axis of the time series
    tau : float
        The averaged persistence

    Returns
    -------

    y : array
        The AR(1) time series

    References
    ----------

    Schulz, M. & Mudelsee, M. REDFIT: estimating red-noise spectra directly from unevenly spaced
        paleoclimatic time series. Computers & Geosciences 28, 421–426 (2002).

    '''
    n = np.size(t)
    y = np.zeros(n)
    y[0] = 0  # initializing

    for i in range(1, n):
        scaled_dt = (t[i] - t[i-1]) / tau
        rho = np.exp(-scaled_dt)
        err = np.random.normal(0, np.sqrt(1 - rho**2)*output_sigma, 1)
        y[i] = y[i-1]*rho + err

    return y


def ar1_fit(y, t=None):
    ''' Return lag-1 autocorrelation
    
    Returns the lag-1 autocorrelation from AR(1) fit OR persistence from tauest.

    Parameters
    ----------

    y : array
        The time series
    t : array
        The time axis of that series

    Returns
    -------

    g : float
        Lag-1 autocorrelation coefficient (for evenly-spaced time series)
        OR estimated persistence (for unevenly-spaced time series)

    See also
    --------

    pyleoclim.utils.tsbase.is_evenly_spaced : Check if a time axis is evenly spaced, within a given tolerance

    pyleoclim.utils.tsmodel.tau_estimation : Estimates the  temporal decay scale of an (un)evenly spaced time series.

    '''

    if is_evenly_spaced(t):
        g = ar1_fit_evenly(y)
    else:
        #  g = tau_estimation(y, t, detrend=detrend, params=params)
        g = tau_estimation(y, t)

    return g

def ar1_sim(y, p, t=None):
    '''Simulate AR(1) process(es) with sample autocorrelation value

    Produce p realizations of an AR(1) process of length n with lag-1 autocorrelation g calculated from `y` and (if provided) `t`

    Parameters
    ----------

    y : array
        a time series; NaNs not allowed
    p : int
        column dimension (number of surrogates)
    t : array
        the time axis of the series

    Returns
    -------

    ysim : array
        n by p matrix of simulated AR(1) vector

    See also
    --------

    pyleoclim.utils.tsmodel.ar1_model : Simulates a (possibly irregularly-sampled) AR(1) process with given decay constant tau, à la REDFIT.

    pyleoclim.utils.tsmodel.ar1_fit : Returns the lag-1 autocorrelation from AR(1) fit OR persistence from tauest.

    pyleoclim.utils.tsmodel.ar1_fit_evenly : Returns the lag-1 autocorrelation from AR(1) fit assuming even temporal spacing.

    pyleoclim.utils.tsmodel.tau_estimation : Estimates the  temporal decay scale of an (un)evenly spaced time series.

    '''
    n = np.size(y)
    ysim = np.empty(shape=(n, p))  # declare array

    sig = np.std(y)
    if is_evenly_spaced(t):
        g = ar1_fit_evenly(y)

        # specify model parameters (statmodel want lag0 coefficent as unity)
        ar = np.r_[1, -g]  # AR model parameter
        ma = np.r_[1, 0.0]  # MA model parameters
        sig_n = sig*np.sqrt(1-g**2)  # theoretical noise variance for the process to achieve the same variance as y

        # simulate AR(1) model for each column
        for i in np.arange(p):
            ysim[:, i] = arma_generate_sample(ar, ma, nsample=n, scale=sig_n, burnin=50) # statsmodels v0.12+
    else:
        #  tau_est = ar1_fit(y, t=t, detrend=detrend, params=params)
        tau_est = tau_estimation(y, t)
        for i in np.arange(p):
            # the output of ar1_model has unit variance,
            # multiply by sig to be consistent with the original input timeseries
            ysim[:, i] = ar1_model(t, tau_est, output_sigma=sig)

    if p == 1:
        ysim = ysim[:, 0]

    return ysim

def gen_ar1_evenly(t, g, scale=1, burnin=50):
    ''' Generate AR(1) series samples

    Wrapper for the function [statsmodels.tsa.arima_process.arma_generate_sample](https://www.statsmodels.org/stable/generated/statsmodels.tsa.arima_process.arma_generate_sample.html)
    used to generate an ARMA

    Parameters
    ----------

    t : array
        the time axis
    
    g : float
        lag-1 autocorrelation

    scale : float
        The standard deviation of noise.

    burnin : int
        Number of observation at the beginning of the sample to drop. Used to reduce dependence on initial values.

    Returns
    -------
    y : array
        the generated AR(1) series

    See also
    --------

    pyleoclim.utils.tsmodel.gen_ts : Generate pyleoclim.Series with timeseries models

    '''
    ar = np.r_[1, -g]  # AR model parameter
    ma = np.r_[1, 0.0]  # MA model parameters
    y = arma_generate_sample(ar, ma, nsample=np.size(t), scale=scale, burnin=burnin)
    return y


def ar1_fit_evenly(y):
    ''' Returns the lag-1 autocorrelation from AR(1) fit.

    Uses [statsmodels.tsa.arima.model.ARIMA](https://www.statsmodels.org/devel/generated/statsmodels.tsa.arima.model.ARIMA.html) to
    calculate lag-1 autocorrelation

    Parameters
    ----------
    y : array
        Vector of (float) numbers as a time series

    Returns
    -------
    g : float
        Lag-1 autocorrelation coefficient

    '''
    # syntax compatible with statsmodels v0.11.1
    #ar1_mod = sm.tsa.ARMA(y, (1, 0), missing='drop').fit(trend='nc', disp=0)
    # syntax compatible with statsmodels v0.12
    ar1_mod = ARIMA(y, order = (1, 0, 0), missing='drop',trend='ct').fit()
    g = ar1_mod.params[2]

    if g > 1:
        print('Warning: AR(1) fitted autocorrelation greater than 1; setting to 1-eps^{1/4}')
        eps = np.spacing(1.0)
        g = 1.0 - eps**(1/4)

    return g

def tau_estimation(y, t):
#  def tau_estimation(y, t, detrend=False, params=["default", 4, 0, 1], gaussianize=False, standardize=True):
    ''' Estimates the  temporal decay scale of an (un)evenly spaced time series.

    Esimtates the temporal decay scale of an (un)evenly spaced time series. 
    Uses [scipy.optimize.minimize_scalar](https://docs.scipy.org/doc/scipy/reference/optimize.minimize_scalar-bounded.html)

    Parameters
    ----------

    y : array
        A time series
    t : array
        Time axis of the time series

    Returns
    -------

    tau_est : float
        The estimated persistence

    References
    ----------

    Mudelsee, M. TAUEST: A Computer Program for Estimating Persistence in Unevenly Spaced Weather/Climate Time Series.
        Comput. Geosci. 28, 69–72 (2002).

    '''
    #  pd_y = preprocess(y, t, detrend=detrend, params=params, gaussianize=gaussianize, standardize=standardize)
    dt = np.diff(t)
    #  assert dt > 0, "The time point should be increasing!"

    def ar1_fun(a):
        #  return np.sum((pd_y[1:] - pd_y[:-1]*a**dt)**2)
        return np.sum((y[1:] - y[:-1]*a**dt)**2)

    a_est = optimize.minimize_scalar(ar1_fun, bounds=[0, 1], method='bounded').x
    #  a_est = optimize.minimize_scalar(ar1_fun, method='brent').x

    tau_est = -1 / np.log(a_est)

    return tau_est


def colored_noise(alpha, t, f0=None, m=None, seed=None):
    ''' Generate a colored noise timeseries

    Parameters
    ----------
    alpha : float
        exponent of the 1/f^alpha noise

    t : float
        time vector of the generated noise

    f0 : float
        fundamental frequency

    m : int
        maximum number of the waves, which determines the highest frequency of the components in the synthetic noise

    Returns
    -------

    y : array
        the generated 1/f^alpha noise

    See also
    --------

    pyleoclim.utils.tsmodel.gen_ts : Generate pyleoclim.Series with timeseries models

    References
    ----------

    Eq. (15) in Kirchner, J. W. Aliasing in 1/f(alpha) noise spectra: origins, consequences, and remedies.
        Phys Rev E Stat Nonlin Soft Matter Phys 71, 066110 (2005).

    '''
    n = np.size(t)  # number of time points
    y = np.zeros(n)

    if f0 is None:
        f0 = 1/n  # fundamental frequency
    if m is None:
        m = n//2

    k = np.arange(m) + 1  # wave numbers

    if seed is not None:
        np.random.seed(seed)

    theta = np.random.rand(int(m))*2*np.pi  # random phase
    for j in range(n):
        coeff = (k*f0)**(-alpha/2)
        sin_func = np.sin(2*np.pi*k*f0*t[j] + theta)
        y[j] = np.sum(coeff*sin_func)

    return y

def colored_noise_2regimes(alpha1, alpha2, f_break, t, f0=None, m=None, seed=None):
    ''' Generate a colored noise timeseries with two regimes

    Parameters
    ----------

    alpha1, alpha2 : float
        the exponent of the 1/f^alpha noise

    f_break : float
        the frequency where the scaling breaks

    t : float
        time vector of the generated noise
    f0 : float
        fundamental frequency
    m : int
        maximum number of the waves, which determines the highest frequency of the components in the synthetic noise

    Returns
    -------

    y : array
        the generated 1/f^alpha noise

    See also
    --------

    pyleoclim.utils.tsmodel.gen_ts : Generate pyleoclim.Series with timeseries models

    References
    ----------

     Eq. (15) in Kirchner, J. W. Aliasing in 1/f(alpha) noise spectra: origins, consequences, and remedies.
         Phys Rev E Stat Nonlin Soft Matter Phys 71, 066110 (2005).
    '''
    n = np.size(t)  # number of time points
    y = np.zeros(n)

    if f0 is None:
        f0 = 1/n  # fundamental frequency
    if m is None:
        m = n//2  # so the aliasing is limited

    k = np.arange(m) + 1  # wave numbers

    if seed is not None:
        np.random.seed(seed)

    theta = np.random.rand(int(m))*2*np.pi  # random phase

    f_vec = k*f0
    regime1= k*f0>=f_break
    regime2= k*f0<=f_break
    f_vec1 = f_vec[regime1]
    f_vec2 = f_vec[regime2]
    s = np.exp(alpha1/alpha2*np.log(f_vec1[0])) / f_vec2[-1]

    for j in range(n):
        coeff = np.ndarray((np.size(f_vec)))
        coeff[regime1] = f_vec1**(-alpha1/2)
        coeff[regime2] = (s*f_vec2)**(-alpha2/2)
        sin_func = np.sin(2*np.pi*k*f0*t[j] + theta)
        y[j] = np.sum(coeff*sin_func)

    return y

def gen_ts(model, t=None, nt=1000, **kwargs):
    ''' Generate pyleoclim.Series with timeseries models

    Parameters
    ----------

    model : str, {'colored_noise', 'colored_noise_2regimes', 'ar1'}
        the timeseries model to use
        - colored_noise : colored noise with one scaling slope
        - colored_noise_2regimes : colored noise with two regimes of two different scaling slopes
        - ar1 : AR(1) series

    t : array
        the time axis

    nt : number of time points
        only works if 't' is None, and it will use an evenly-spaced vector with nt points

    kwargs : dict
        the keyward arguments for the specified timeseries model

    Returns
    -------

    t, v : NumPy arrays
        time axis and values

    See also
    --------

    pyleoclim.utils.tsmodel.colored_noise : Generate a colored noise timeseries

    pyleoclim.utils.tsmodel.colored_noise_2regimes : Generate a colored noise timeseries with two regimes

    pyleoclim.utils.tsmodel.gen_ar1_evenly : Generate an AR(1) series

    '''

    if t is None:
        t = np.arange(nt)

    tsm = {
        'colored_noise': colored_noise,
        'colored_noise_2regimes': colored_noise_2regimes,
        'ar1': gen_ar1_evenly,
    }

    tsm_args = {}
    tsm_args['colored_noise'] = {'alpha': 1}
    tsm_args['colored_noise_2regimes'] = {'alpha1': 1/2, 'alpha2': 2, 'f_break': 1/20}
    tsm_args['ar1'] = {'g': 0.5}
    tsm_args[model].update(kwargs)

    v = tsm[model](t=t, **tsm_args[model])
    #ts = Series(time=t, value=v)

    return t, v

# def fBMsim(N=128, H=0.25):
#     '''Simple method to generate fractional Brownian Motion

#     Parameters
#     ----------

#     N : int
#         the length of the simulated time series
#     H : float
#         Hurst index, should be in (0, 1). The relationship between H and the scaling exponent beta is
#         H = (beta-1) / 2

#     Returns
#     -------

#     xfBm : array
#         the simulated fractional Brownian Motion time series

#     References
#     ----------

#     1. http://cours-physique.lps.ens.fr/index.php/TD11_Correlated_Noise_2011
#     2. https://www.wikiwand.com/en/Fractional_Brownian_motion

#     @authors: jeg, fzhu
#     '''
#     assert isinstance(N, int) and N >= 1
#     assert H > 0 and H < 1, "H should be in (0, 1)!"

#     HH = 2 * H

#     ns = N-1  # number of steps
#     covariance = np.ones((ns, ns))

#     for i in range(ns):
#         for j in range(i, ns):
#             x = np.abs(i-j)
#             covariance[i, j] = covariance[j, i] = (np.abs(x-1)**HH + (x+1)**HH - 2*x**HH) / 2.

#     w, v = np.linalg.eig(covariance)

#     A = np.zeros((ns, ns))
#     for i in range(ns):
#         for j in range(i, ns):
#             A[i, j] = A[j, i] = np.sum(np.sqrt(w) * v[i, :] * v[j, :])

#     xi = np.random.randn((ns))
#     eta = np.dot(A, xi)

#     xfBm = np.zeros(N)
#     xfBm[0] = 0
#     for i in range(1, N):
#         xfBm[i] = xfBm[i-1] + eta[i-1]

#     return xfBm
